# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import tempfile
from functools import partial

import numpy as np
import pytest
import torch
from mmcls.datasets.pipelines import Compose
from nncf.torch.nncf_network import NNCFNetwork

from otx.algorithms.common.adapters.mmcv.nncf.utils import (
    get_fake_input,
    model_eval,
    wrap_nncf_model,
)
from otx.algorithms.common.adapters.nncf.compression import NNCFMetaState
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmcv.nncf.test_helpers import (
    create_config,
    create_dataloader,
    create_eval_fn,
    create_model,
)


@e2e_pytest_unit
def test_get_fake_input():
    pipeline = Compose([{"type": "Resize", "size": (50, 50)}, {"type": "Collect", "keys": ["img"]}])

    output = get_fake_input(pipeline)
    assert torch.equal(output["img"][0], torch.zeros((1, 50, 50, 3), dtype=torch.uint8))

    if torch.cuda.is_available():
        output = get_fake_input(pipeline, device="cuda")
        assert torch.equal(
            output["img"][0],
            torch.zeros((1, 50, 50, 3), dtype=torch.uint8, device="cuda"),
        )

    output = get_fake_input(pipeline, np.zeros((128, 128, 3), dtype=np.uint8))
    assert torch.equal(output["img"][0], torch.zeros((1, 50, 50, 3), dtype=torch.uint8))


@e2e_pytest_unit
def test_model_eval():
    mock_model = create_model()
    mock_config = create_config()
    mock_eval_fn = create_eval_fn()
    dataloader = create_dataloader()

    if torch.cuda.is_available():
        model_eval(
            mock_model,
            config=mock_config,
            val_dataloader=dataloader,
            evaluate_fn=mock_eval_fn,
            distributed=False,
        )

        with pytest.raises(RuntimeError):
            model_eval(
                mock_model,
                config=mock_config,
                val_dataloader=None,
                evaluate_fn=mock_eval_fn,
                distributed=False,
            )

        with pytest.raises(RuntimeError):
            mock_config["nncf_config"]["target_metric_name"] = "failed"
            model_eval(
                mock_model,
                config=mock_config,
                val_dataloader=dataloader,
                evaluate_fn=mock_eval_fn,
                distributed=False,
            )


@e2e_pytest_unit
def test_wrap_nncf_model():
    mock_model = create_model()
    mock_config = create_config()
    mock_eval_fn = create_eval_fn()
    dataloader = create_dataloader()

    pipeline = Compose(
        [
            {"type": "Resize", "size": (50, 50)},
            {"type": "Normalize", "mean": [0, 0, 0], "std": [1, 1, 1]},
            {"type": "Collect", "keys": ["img"]},
        ]
    )
    get_fake_input_fn = partial(get_fake_input, pipeline)

    model_eval_fn = partial(
        model_eval,
        config=mock_config,
        val_dataloader=dataloader,
        evaluate_fn=mock_eval_fn,
        distributed=False,
    )

    ctrl, model = wrap_nncf_model(
        mock_config,
        mock_model,
        model_eval_fn=model_eval_fn,
        get_fake_input_fn=get_fake_input_fn,
        dataloader_for_init=dataloader,
        is_accuracy_aware=True,
    )
    assert isinstance(model, NNCFNetwork)

    mock_model = create_model()
    mock_config.nncf_config["input_info"] = {"sample_size": (1, 3, 128, 128)}
    ctrl, model = wrap_nncf_model(
        mock_config,
        mock_model,
        model_eval_fn=model_eval_fn,
        get_fake_input_fn=get_fake_input_fn,
        dataloader_for_init=dataloader,
        is_accuracy_aware=True,
    )
    assert isinstance(model, NNCFNetwork)
    mock_config.nncf_config.pop("input_info")

    with tempfile.TemporaryDirectory() as tempdir:
        mock_model = create_model()
        model_path = os.path.join(tempdir, "model.bin")
        torch.save(mock_model.state_dict(), model_path)
        mock_config.load_from = model_path
        mock_config.nncf_config.log_dir = tempdir
        ctrl, model = wrap_nncf_model(
            mock_config,
            mock_model,
            model_eval_fn=model_eval_fn,
            get_fake_input_fn=get_fake_input_fn,
            dataloader_for_init=dataloader,
            is_accuracy_aware=True,
        )
        assert isinstance(model, NNCFNetwork)
        mock_config.pop("load_from")
        mock_config.nncf_config.pop("log_dir")

        init_state_dict = {
            "meta": {
                "nncf_enable_compression": True,
                "nncf_meta": NNCFMetaState(compression_ctrl=ctrl.get_compression_state()),
            },
            "state_dict": model.state_dict(),
        }
        mock_model = create_model()
        ctrl, model = wrap_nncf_model(
            mock_config,
            mock_model,
            model_eval_fn=model_eval_fn,
            get_fake_input_fn=get_fake_input_fn,
            dataloader_for_init=dataloader,
            init_state_dict=init_state_dict,
            is_accuracy_aware=True,
        )



"""OTX searching command 'otx find'.

Through this command, you can check the tasks, templates, and backbones available in OTX.
"""
# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import os
from textwrap import fill

from prettytable import PrettyTable

from otx.cli.registry import Registry
from otx.cli.utils.importing import SUPPORTED_BACKBONE_BACKENDS, get_otx_root_path

# pylint: disable=too-many-locals

SUPPORTED_TASKS = (
    "CLASSIFICATION",
    "DETECTION",
    "ROTATED_DETECTION",
    "INSTANCE_SEGMENTATION",
    "SEGMENTATION",
    "ACTION_CLASSIFICATION",
    "ACTION_DETECTION",
    "ANOMALY_CLASSIFICATION",
    "ANOMALY_DETECTION",
    "ANOMALY_SEGMENTATION",
)


def parse_args():
    """Parses command line arguments."""

    parser = argparse.ArgumentParser()
    parser.add_argument("--task", help=f"The currently supported options: {SUPPORTED_TASKS}.")
    parser.add_argument(
        "--template", action="store_true", help="Shows a list of templates that can be used immediately."
    )
    parser.add_argument(
        "--backbone",
        nargs="+",
        help=f"The currently supported options: {list(SUPPORTED_BACKBONE_BACKENDS.keys())}.",
    )

    return parser.parse_args()


def generate_backbone_rows(index: int, backbone_type: str, meta_data: dict):
    """Generate table row for backbone json format.

    It expects a json file format from otx/cli/builder/supported_backbone.
    index: The index of each backbone (int)
    backbone_type: The backbone type want to add (str)
    meta_data: This is the metadata of the backbone type (dict)
        Metadata keys expect required, options, and available.
    """
    max_row_width = 40
    rows = []
    required_args = meta_data["required"] if meta_data["required"] else [""]
    required_options = meta_data["options"]
    use_backbone_type = False
    for arg in required_args:
        row = []
        options = list(map(str, required_options[arg])) if arg in required_options else []
        if use_backbone_type:
            row.append("")  # Index
            row.append("")  # backbone_type
        else:
            row.append(str(index))  # Index
            row.append(backbone_type)  # backbone_type
            use_backbone_type = True
        row.append(arg)  # Required-Args
        option_str = ", ".join(options) if options else ""
        row.append(fill(option_str, width=max_row_width))  # Options
        rows.append(row)
    return rows


def main():
    """Main function for model templates & backbone searching.

    When the template argment is input, the templates based on the otx folder are displayed.
    Given a backbone argument as input,
    it displays a list of backbones available in the backend of the relevant task.
    """

    args = parse_args()

    otx_root = get_otx_root_path()
    otx_registry = Registry(otx_root).filter(task_type=args.task)

    if not args.backbone or args.template:
        template_table = PrettyTable(["TASK", "ID", "NAME", "BASE PATH"])
        for template in otx_registry.templates:
            relpath = os.path.relpath(template.model_template_path, os.path.abspath("."))
            template_table.add_row(
                [
                    template.task_type,
                    template.model_template_id,
                    template.name,
                    relpath,
                ]
            )
        print(template_table)

    if args.backbone:
        all_backbones = otx_registry.get_backbones(args.backbone)
        backbone_table = PrettyTable(["Index", "Backbone Type", "Required-Args", "Options"])
        row_index = 1
        for _, backbone_meta in all_backbones.items():
            for backbone_type, meta_data in backbone_meta.items():
                rows = generate_backbone_rows(row_index, backbone_type, meta_data)
                backbone_table.add_rows(rows)
                row_index += 1
        print(backbone_table)

    return dict(retcode=0, task_type=args.task)


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmdet.models.builder import ROI_EXTRACTORS
from mmdet.models.roi_heads.roi_extractors.single_level_roi_extractor import (
    SingleRoIExtractor as OriginSingleRoIExtractor,
)
from torch import nn
from torch.nn import functional as F
from torch.nn.modules.utils import _pair


@ROI_EXTRACTORS.register_module(force=True)
class SingleRoIExtractor(OriginSingleRoIExtractor):
    def build_roi_layers(self, layer_cfg, featmap_strides):

        if layer_cfg["type"] == "RoIInterpolationPool":
            cfg = layer_cfg.copy()
            cfg.pop("type")
            return nn.ModuleList([RoIInterpolationPool(spatial_scale=1 / s, **cfg) for s in featmap_strides])
        else:
            return super().build_roi_layers(layer_cfg, featmap_strides)


class RoIInterpolationPool(nn.Module):
    def __init__(self, output_size, spatial_scale, mode="bilinear"):
        super().__init__()

        self.output_size = _pair(output_size)
        self.spatial_scale = float(spatial_scale)
        self.mode = mode

    def forward(self, input, rois):
        outs = []
        for roi in rois:
            batch_idx = roi[0].to(dtype=torch.int)
            roi = roi[1:]
            roi = roi * self.spatial_scale
            x1, y1 = roi[:2].floor().to(dtype=torch.int)
            x2, y2 = roi[2:].ceil().to(dtype=torch.int)
            outs.append(
                F.interpolate(
                    input[batch_idx : batch_idx + 1, :, y1:y2, x1:x2],
                    self.output_size,
                    mode=self.mode,
                    align_corners=True,
                )
            )
        outs = torch.cat(outs, 0)
        return outs

    def __repr__(self):
        s = self.__class__.__name__
        s += f"(output_size={self.output_size}, "
        s += f"spatial_scale={self.spatial_scale}, "
        s += f"mode={self.mode})"
        return s



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


class SAMClassifierMixin(object):
    """SAM-enabled BaseClassifier mix-in"""

    def train_step(self, data, optimizer=None, **kwargs):
        # Saving current batch data to compute SAM gradient
        # Rest of SAM logics are implented in SAMOptimizerHook
        self.current_batch = data

        return super().train_step(data, optimizer, **kwargs)



"""Utils for OTX Detection."""
# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import colorsys
import random
from typing import Optional, Sequence

import numpy as np

from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.utils.argument_checks import check_input_parameters_type

# pylint: disable=invalid-name


class ColorPalette:
    """ColorPalette class."""

    @check_input_parameters_type()
    def __init__(self, n: int, rng: Optional[random.Random] = None):
        assert n > 0

        if rng is None:
            rng = random.Random(0xACE)

        candidates_num = 100
        hsv_colors = [(1.0, 1.0, 1.0)]
        for _ in range(1, n):
            colors_candidates = [
                (rng.random(), rng.uniform(0.8, 1.0), rng.uniform(0.5, 1.0)) for _ in range(candidates_num)
            ]
            min_distances = [self._min_distance(hsv_colors, c) for c in colors_candidates]
            arg_max = np.argmax(min_distances)
            hsv_colors.append(colors_candidates[arg_max])

        self.palette = [Color(*self._hsv2rgb(*hsv)) for hsv in hsv_colors]

    @staticmethod
    def _dist(c1, c2):
        dh = min(abs(c1[0] - c2[0]), 1 - abs(c1[0] - c2[0])) * 2
        ds = abs(c1[1] - c2[1])
        dv = abs(c1[2] - c2[2])
        return dh * dh + ds * ds + dv * dv

    @classmethod
    def _min_distance(cls, colors_set, color_candidate):
        distances = [cls._dist(o, color_candidate) for o in colors_set]
        return np.min(distances)

    @staticmethod
    def _hsv2rgb(h, s, v):
        return tuple(round(c * 255) for c in colorsys.hsv_to_rgb(h, s, v))

    @check_input_parameters_type()
    def __getitem__(self, n: int):
        """Return item from index function ColorPalette."""
        return self.palette[n % len(self.palette)]

    def __len__(self):
        """Return length of ColorPalette."""
        return len(self.palette)


@check_input_parameters_type()
def generate_label_schema(label_names: Sequence[str], label_domain: Domain = Domain.DETECTION):
    """Generating label_schema function."""
    colors = ColorPalette(len(label_names)) if len(label_names) > 0 else []
    not_empty_labels = [
        LabelEntity(name=name, color=colors[i], domain=label_domain, id=ID(f"{i:08}"))
        for i, name in enumerate(label_names)
    ]
    emptylabel = LabelEntity(
        name="Empty label",
        color=Color(42, 43, 46),
        is_empty=True,
        domain=label_domain,
        id=ID(f"{len(not_empty_labels):08}"),
    )

    label_schema = LabelSchemaEntity()
    exclusive_group = LabelGroup(name="labels", labels=not_empty_labels, group_type=LabelGroupType.EXCLUSIVE)
    empty_group = LabelGroup(name="empty", labels=[emptylabel], group_type=LabelGroupType.EMPTY_LABEL)
    label_schema.add_group(exclusive_group)
    label_schema.add_group(empty_group)
    return label_schema



"""Test utils."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import pytest
import torch

from otx.mpa.modules.models.losses.utils import LossEqualizer


class TestLossEqualizer:
    """Test loss equalizer."""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Create the loss object"""
        weights = {"loss_a": 1.0, "loss_b": 2.0, "loss_c": 3.0}
        self.loss_equalizer = LossEqualizer(weights, momentum=0)

    def test_loss_equalizer(self):
        """Test value"""
        losses = {
            "loss_a": torch.tensor(10.0),
            "loss_b": torch.tensor(4.0),
            "loss_c": torch.tensor(1.0),
        }
        result = {val.item() for val in self.loss_equalizer.reweight(losses).values()}
        assert result == {2.5, 5.0, 7.5}



_base_ = ["./train.py", "../_base_/data/coco_otx.py", "../_base_/models/detectors/detector.py"]

task_adapt = dict(
    type="mpa",
    op="REPLACE",
    efficient_mode=False,
)

runner = dict(max_epochs=30)

evaluation = dict(interval=1, metric="mAP", save_best="mAP")

custom_hooks = [
    dict(
        type="LazyEarlyStoppingHook",
        start=3,
        patience=10,
        iteration_patience=0,
        metric="mAP",
        interval=1,
        priority=75,
    ),
]

lr_config = dict(
    policy="ReduceLROnPlateau",
    metric="mAP",
    patience=5,
    iteration_patience=0,
    interval=1,
    min_lr=1e-06,
    warmup="linear",
    warmup_iters=200,
    warmup_ratio=0.3333333333333333,
)

ignore = True
adaptive_validation_interval = dict(
    max_interval=5,
    enable_adaptive_interval_hook=True,
    enable_eval_before_run=True,
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os.path as osp
from contextlib import nullcontext

import numpy as np
import torch
from mmcls.datasets import build_dataloader as mmcls_build_dataloader
from mmcls.datasets import build_dataset as mmcls_build_dataset
from mmcv import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    build_data_parallel,
    build_dataloader,
    build_dataset,
)
from otx.mpa.modules.hooks.recording_forward_hooks import (
    FeatureVectorHook,
    ReciproCAMHook,
)
from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import ClsStage

logger = get_logger()


@STAGES.register_module()
class ClsInferrer(ClsStage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dataset = None

    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):
        """Run inference stage for classification

        - Configuration
        - Environment setup
        - Run inference via MMClassification -> MMCV
        """
        self._init_logger()
        mode = kwargs.get("mode", "train")
        if mode not in self.mode:
            logger.warning(f"Supported modes are {self.mode} but '{mode}' is given.")
            return {}

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs)
        logger.info("infer!")

        model_builder = kwargs.get("model_builder", None)
        dump_features = kwargs.get("dump_features", False)
        dump_saliency_map = kwargs.get("dump_saliency_map", False)
        eval = kwargs.get("eval", False)
        outputs = self.infer(
            cfg,
            model_builder=model_builder,
            eval=eval,
            dump_features=dump_features,
            dump_saliency_map=dump_saliency_map,
        )

        return dict(outputs=outputs)

    def infer(self, cfg, model_builder=None, eval=False, dump_features=False, dump_saliency_map=False):
        # TODO: distributed inference

        if cfg.get("task_adapt", False) and not eval:
            data_cfg = cfg.data.train.copy()
            data_cfg.pipeline = cfg.data.test.pipeline
        else:
            data_cfg = cfg.data.test.copy()

        data_cfg = Config(
            ConfigDict(
                data=ConfigDict(
                    samples_per_gpu=cfg.data.get("samples_per_gpu", 1),
                    workers_per_gpu=cfg.data.get("workers_per_gpu", 0),
                    test=data_cfg,
                    test_dataloader=cfg.data.get("test_dataloader", {}).copy(),
                ),
                gpu_ids=cfg.gpu_ids,
                seed=cfg.get("seed", None),
                model_task=cfg.model_task,
            )
        )
        self.configure_samples_per_gpu(data_cfg, "test", distributed=False)
        self.configure_compat_cfg(data_cfg)

        # Data loader
        self.dataset = build_dataset(data_cfg, "test", mmcls_build_dataset)
        test_dataloader = build_dataloader(
            self.dataset,
            data_cfg,
            "test",
            mmcls_build_dataloader,
            distributed=False,
            round_up=False,
        )

        # Model
        model = self.build_model(cfg, model_builder, fp16=cfg.get("fp16", False))
        model.eval()
        feature_model = self._get_feature_module(model)
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        self.set_inference_progress_callback(model, cfg)

        eval_predictions = []
        feature_vectors = []
        saliency_maps = []
        with FeatureVectorHook(feature_model) if dump_features else nullcontext() as feature_vector_hook:
            with ReciproCAMHook(feature_model) if dump_saliency_map else nullcontext() as forward_explainer_hook:
                for data in test_dataloader:
                    with torch.no_grad():
                        result = model(return_loss=False, **data)
                    eval_predictions.extend(result)
                feature_vectors = feature_vector_hook.records if dump_features else [None] * len(self.dataset)
                saliency_maps = forward_explainer_hook.records if dump_saliency_map else [None] * len(self.dataset)
                if len(eval_predictions) == 0:
                    eval_predictions = [None] * len(self.dataset)

        assert len(eval_predictions) == len(feature_vectors) == len(saliency_maps), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(eval_predictions)}, {len(feature_vectors)}, and {len(saliency_maps)}"
        )

        outputs = dict(
            eval_predictions=eval_predictions,
            feature_vectors=feature_vectors,
            saliency_maps=saliency_maps,
        )
        return outputs



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from contextlib import nullcontext

import torch
from mmcv.utils import Config, ConfigDict
from mmdet.apis import single_gpu_test
from mmdet.datasets import build_dataloader as mmdet_build_dataloader
from mmdet.datasets import build_dataset as mmdet_build_dataset
from mmdet.datasets import replace_ImageToTensor
from mmdet.models.detectors import TwoStageDetector

from otx.algorithms.common.adapters.mmcv.utils import (
    build_data_parallel,
    build_dataloader,
    build_dataset,
)
from otx.algorithms.detection.adapters.mmdet.data import ImageTilingDataset
from otx.mpa.modules.hooks.recording_forward_hooks import (
    ActivationMapHook,
    DetSaliencyMapHook,
    FeatureVectorHook,
)
from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import DetectionStage

logger = get_logger()


@STAGES.register_module()
class DetectionInferrer(DetectionStage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dataset = None

    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):
        """Run inference stage for detection

        - Configuration
        - Environment setup
        - Run inference via MMDetection -> MMCV
        """
        self._init_logger()
        mode = kwargs.get("mode", "train")
        if mode not in self.mode:
            logger.warning(f"Supported modes are {self.mode} but '{mode}' is given.")
            return {}

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs)
        logger.info("infer!")

        model_builder = kwargs.get("model_builder", None)
        dump_features = kwargs.get("dump_features", False)
        dump_saliency_map = kwargs.get("dump_saliency_map", False)
        eval = kwargs.get("eval", False)
        outputs = self.infer(
            cfg,
            model_builder=model_builder,
            eval=eval,
            dump_features=dump_features,
            dump_saliency_map=dump_saliency_map,
        )

        # Save outputs
        # output_file_path = osp.join(cfg.work_dir, 'infer_result.npy')
        # np.save(output_file_path, outputs, allow_pickle=True)
        return dict(
            # output_file_path=output_file_path,
            outputs=outputs
        )
        # TODO: save in json
        """
        class NumpyEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, np.ndarray):
                    return obj.tolist()
                return json.JSONEncoder.default(self, obj)

        a = np.array([[1, 2, 3], [4, 5, 6]])
        json_dump = json.dumps({'a': a, 'aa': [2, (2, 3, 4), a], 'bb': [2]},
                               cls=NumpyEncoder)
        print(json_dump)
        """

    def infer(self, cfg, model_builder=None, eval=False, dump_features=False, dump_saliency_map=False):
        # TODO: distributed inference

        data_cfg = cfg.data.test.copy()

        # Input source
        if "input_source" in cfg:
            input_source = cfg.get("input_source")
            logger.info(f"Inferring on input source: data.{input_source}")
            if input_source == "train":
                src_data_cfg = self.get_data_cfg(cfg, input_source)
            else:
                src_data_cfg = cfg.data[input_source]
            data_cfg.test_mode = src_data_cfg.get("test_mode", False)
            data_cfg.ann_file = src_data_cfg.ann_file
            data_cfg.img_prefix = src_data_cfg.img_prefix
            if "classes" in src_data_cfg:
                data_cfg.classes = src_data_cfg.classes

        data_cfg = Config(
            ConfigDict(
                data=ConfigDict(
                    samples_per_gpu=cfg.data.get("samples_per_gpu", 1),
                    workers_per_gpu=cfg.data.get("workers_per_gpu", 0),
                    test=data_cfg,
                    test_dataloader=cfg.data.get("test_dataloader", {}).copy(),
                ),
                gpu_ids=cfg.gpu_ids,
                seed=cfg.get("seed", None),
                model_task=cfg.model_task,
            )
        )
        self.configure_samples_per_gpu(data_cfg, "test", distributed=False)
        self.configure_compat_cfg(data_cfg)
        samples_per_gpu = data_cfg.data.test_dataloader.get("samples_per_gpu", 1)
        if samples_per_gpu > 1:
            # Replace 'ImageToTensor' to 'DefaultFormatBundle'
            data_cfg.data.test.pipeline = replace_ImageToTensor(data_cfg.data.test.pipeline)

        # Data loader
        self.dataset = build_dataset(data_cfg, "test", mmdet_build_dataset)
        test_dataloader = build_dataloader(
            self.dataset,
            data_cfg,
            "test",
            mmdet_build_dataloader,
            distributed=False,
        )

        # Target classes
        if "task_adapt" in cfg:
            target_classes = cfg.task_adapt.final
            if len(target_classes) < 1:
                raise KeyError(
                    f"target_classes={target_classes} is empty check the metadata from model ckpt or recipe "
                    "configuration"
                )
        else:
            target_classes = self.dataset.CLASSES

        # Model
        cfg.model.pretrained = None
        if cfg.model.get("neck"):
            if isinstance(cfg.model.neck, list):
                for neck_cfg in cfg.model.neck:
                    if neck_cfg.get("rfp_backbone"):
                        if neck_cfg.rfp_backbone.get("pretrained"):
                            neck_cfg.rfp_backbone.pretrained = None
            elif cfg.model.neck.get("rfp_backbone"):
                if cfg.model.neck.rfp_backbone.get("pretrained"):
                    cfg.model.neck.rfp_backbone.pretrained = None
        # TODO: Check Inference FP16 Support
        model = self.build_model(cfg, model_builder, fp16=False)
        model.CLASSES = target_classes
        model.eval()
        feature_model = self._get_feature_module(model)
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        self.set_inference_progress_callback(model, cfg)

        # Class-wise Saliency map for Single-Stage Detector, otherwise use class-ignore saliency map.
        if not dump_saliency_map:
            saliency_hook = nullcontext()
        else:
            raw_model = feature_model
            if raw_model.__class__.__name__ == "NNCFNetwork":
                raw_model = raw_model.get_nncf_wrapped_model()
            if isinstance(raw_model, TwoStageDetector):
                saliency_hook = ActivationMapHook(feature_model)
            else:
                saliency_hook = DetSaliencyMapHook(feature_model)

        eval_predictions = []
        with FeatureVectorHook(feature_model) if dump_features else nullcontext() as feature_vector_hook:
            with saliency_hook:
                eval_predictions = single_gpu_test(model, test_dataloader)
                feature_vectors = feature_vector_hook.records if dump_features else [None] * len(self.dataset)
                saliency_maps = saliency_hook.records if dump_saliency_map else [None] * len(self.dataset)

        for key in ["interval", "tmpdir", "start", "gpu_collect", "save_best", "rule", "dynamic_intervals"]:
            cfg.evaluation.pop(key, None)

        metric = None
        if eval:
            metric = self.dataset.evaluate(eval_predictions, **cfg.evaluation)
            metric = metric["mAP"] if isinstance(cfg.evaluation.metric, list) else metric[cfg.evaluation.metric]

        # Check and unwrap ImageTilingDataset object from TaskAdaptEvalDataset
        dataset = self.dataset
        while hasattr(dataset, "dataset") and not isinstance(dataset, ImageTilingDataset):
            dataset = dataset.dataset

        if isinstance(dataset, ImageTilingDataset):
            feature_vectors = [feature_vectors[i] for i in range(dataset.num_samples)]
            saliency_maps = [saliency_maps[i] for i in range(dataset.num_samples)]
            if not dataset.merged_results:
                eval_predictions = dataset.merge(eval_predictions)
            else:
                eval_predictions = dataset.merged_results

        assert len(eval_predictions) == len(feature_vectors) == len(saliency_maps), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(eval_predictions)}, {len(feature_vectors)}, and {len(saliency_maps)}"
        )

        outputs = dict(
            classes=target_classes,
            detections=eval_predictions,
            metric=metric,
            feature_vectors=feature_vectors,
            saliency_maps=saliency_maps,
        )
        return outputs



"""Collection of utils for task implementation in Action Task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from collections import defaultdict
from typing import List, Union

from mmcv.utils import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    get_data_cfg,
    patch_data_pipeline,
    prepare_work_dir,
)
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.model_template import TaskType
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)


@check_input_parameters_type()
def patch_config(config: Config, data_pipeline_path: str, work_dir: str, task_type: TaskType):
    """Patch recipe config suitable to mmaction."""
    # FIXME omnisource is hard coded
    config.omnisource = None
    config.work_dir = work_dir
    patch_data_pipeline(config, data_pipeline_path)
    if task_type == TaskType.ACTION_CLASSIFICATION:
        _patch_cls_datasets(config)
    elif task_type == TaskType.ACTION_DETECTION:
        _patch_det_dataset(config)
    else:
        raise NotImplementedError(f"{task_type} is not supported in action task")


@check_input_parameters_type()
def _patch_cls_datasets(config: Config):
    """Patch cls dataset config suitable to mmaction."""

    assert "data" in config
    for subset in ("train", "val", "test", "unlabeled"):
        cfg = config.data.get(subset, None)
        if not cfg:
            continue
        cfg.type = "OTXActionClsDataset"
        cfg.otx_dataset = None
        cfg.labels = None


@check_input_parameters_type()
def _patch_det_dataset(config: Config):
    """Patch det dataset config suitable to mmaction."""
    assert "data" in config
    for subset in ("train", "val", "test", "unlabeled"):
        cfg = config.data.get(subset, None)
        if not cfg:
            continue
        cfg.type = "OTXActionDetDataset"


@check_input_parameters_type()
def set_data_classes(config: Config, labels: List[LabelEntity], task_type: TaskType):
    """Setter data classes into config."""
    for subset in ("train", "val", "test"):
        cfg = get_data_cfg(config, subset)
        cfg.labels = labels

    # FIXME classification head name is hard-coded
    if task_type == TaskType.ACTION_CLASSIFICATION:
        config.model["cls_head"].num_classes = len(labels)
    elif task_type == TaskType.ACTION_DETECTION:
        config.model["roi_head"]["bbox_head"].num_classes = len(labels) + 1
        if len(labels) < 5:
            config.model["roi_head"]["bbox_head"]["topk"] = len(labels) - 1


@check_input_parameters_type({"train_dataset": DatasetParamTypeCheck, "val_dataset": DatasetParamTypeCheck})
def prepare_for_training(
    config: Union[Config, ConfigDict],
    train_dataset: DatasetEntity,
    val_dataset: DatasetEntity,
    time_monitor: TimeMonitorCallback,
    learning_curves: defaultdict,
) -> Config:
    """Prepare configs for training phase."""
    prepare_work_dir(config)
    data_train = get_data_cfg(config)
    data_train.otx_dataset = train_dataset
    config.data.val.otx_dataset = val_dataset
    config.custom_hooks.append({"type": "OTXProgressHook", "time_monitor": time_monitor, "verbose": True})
    config.log_config.hooks.append({"type": "OTXLoggerHook", "curves": learning_curves})

    return config



"""A file for a function build_data_parallel()."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# NOTE: a workaround for https://github.com/python/mypy/issues/5028

import os
from typing import Literal, Union, overload

import torch
from mmcv import Config
from mmcv.parallel import MMDataParallel, MMDistributedDataParallel

from otx.api.utils.argument_checks import check_input_parameters_type


@overload
def build_data_parallel(
    model: torch.nn.Module,
    config: Config,
    *,
    distributed: Literal[True],
) -> MMDistributedDataParallel:
    ...


@overload
def build_data_parallel(
    model: torch.nn.Module,
    config: Config,
    *,
    distributed: Literal[False] = False,
) -> MMDataParallel:
    ...


@overload
def build_data_parallel(
    model: torch.nn.Module,
    config: Config,
    *,
    distributed: bool,
) -> Union[MMDataParallel, MMDistributedDataParallel]:
    ...


@check_input_parameters_type()
def build_data_parallel(
    model: torch.nn.Module,
    config: Config,
    *,
    distributed: bool = False,
) -> Union[MMDataParallel, MMDistributedDataParallel]:
    """Prepare model for execution.

    Return MMDataParallel or MMDistributedDataParallel model.

    :param model: Model.
    :param config: config.
    :param distributed: Enable distributed training mode.
    :return:
    """
    if torch.cuda.is_available() and config.get("gpu_ids", []):
        if distributed:
            model = model.cuda()
            # put model on gpus
            find_unused_parameters = config.get("find_unused_parameters", False)
            # Sets the `find_unused_parameters` parameter in
            # torch.nn.parallel.DistributedDataParallel
            model = MMDistributedDataParallel(
                model,
                device_ids=[int(os.environ["LOCAL_RANK"])],
                broadcast_buffers=False,
                find_unused_parameters=find_unused_parameters,
            )
        else:
            model = model.cuda(config.gpu_ids[0])
            model = MMDataParallel(model, device_ids=config.gpu_ids)
    else:
        # temporarily disable cuda for cpu data parallel
        bak = torch.cuda.is_available
        setattr(torch.cuda, "is_available", lambda: False)
        model = MMDataParallel(model, device_ids=[])
        torch.cuda.is_available = bak
    return model



""".This module contains the mapper for label related entities."""
#
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import json
from typing import Dict, Union, cast

from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import (
    LabelGraph,
    LabelGroup,
    LabelGroupType,
    LabelSchemaEntity,
    LabelTree,
)

from .datetime_mapper import DatetimeMapper
from .id_mapper import IDMapper


class ColorMapper:
    """This class maps a `Color` entity to a serialized dictionary, and vice versa."""

    @staticmethod
    def forward(instance: Color) -> dict:
        """Serializes to dict."""

        return {
            "red": instance.red,
            "green": instance.green,
            "blue": instance.blue,
            "alpha": instance.alpha,
        }

    @staticmethod
    def backward(instance: dict) -> Color:
        """Deserializes from dict."""

        return Color(instance["red"], instance["green"], instance["blue"], instance["alpha"])


class LabelMapper:
    """This class maps a `Label` entity to a serialized dictionary, and vice versa."""

    @staticmethod
    def forward(
        instance: LabelEntity,
    ) -> dict:
        """Serializes to dict."""

        return {
            "_id": IDMapper().forward(instance.id_),
            "name": instance.name,
            "color": ColorMapper().forward(instance.color),
            "hotkey": instance.hotkey,
            "domain": str(instance.domain),
            "creation_date": DatetimeMapper.forward(instance.creation_date),
            "is_empty": instance.is_empty,
            "is_anomalous": instance.is_anomalous,
        }

    @staticmethod
    def backward(instance: dict) -> LabelEntity:
        """Deserializes from dict."""

        label_id = IDMapper().backward(instance["_id"])

        domain = str(instance.get("domain"))
        label_domain = Domain[domain]

        label = LabelEntity(
            id=label_id,
            name=instance["name"],
            color=ColorMapper().backward(instance["color"]),
            hotkey=instance.get("hotkey", ""),
            domain=label_domain,
            creation_date=DatetimeMapper.backward(instance["creation_date"]),
            is_empty=instance.get("is_empty", False),
            is_anomalous=instance.get("is_anomalous", False),
        )
        return label


class LabelGroupMapper:
    """This class maps a `LabelGroup` entity to a serialized dictionary, and vice versa."""

    @staticmethod
    def forward(instance: LabelGroup) -> dict:
        """Serializes to dict."""

        return {
            "_id": IDMapper().forward(instance.id_),
            "name": instance.name,
            "label_ids": [IDMapper().forward(label.id_) for label in instance.labels],
            "relation_type": instance.group_type.name,
        }

    @staticmethod
    def backward(instance: dict, all_labels: Dict[ID, LabelEntity]) -> LabelGroup:
        """Deserializes from dict."""

        return LabelGroup(
            id=IDMapper().backward(instance["_id"]),
            name=instance["name"],
            group_type=LabelGroupType[instance["relation_type"]],
            labels=[all_labels[IDMapper().backward(label_id)] for label_id in instance["label_ids"]],
        )


class LabelGraphMapper:
    """This class maps a `LabelGraph` entity to a serialized dictionary, and vice versa."""

    @staticmethod
    def forward(instance: Union[LabelGraph, LabelTree]) -> dict:
        """Serializes to dict."""

        return {
            "type": instance.type,
            "directed": instance.directed,
            "nodes": [IDMapper().forward(label.id_) for label in instance.nodes],
            "edges": [(IDMapper().forward(edge[0].id_), IDMapper().forward(edge[1].id_)) for edge in instance.edges],
        }

    @staticmethod
    def backward(instance: dict, all_labels: Dict[ID, LabelEntity]) -> Union[LabelTree, LabelGraph]:
        """Deserializes from dict."""

        output: Union[LabelTree, LabelGraph]

        instance_type = instance["type"]
        if instance_type == "tree":
            output = LabelTree()
        elif instance_type == "graph":
            output = LabelGraph(instance["directed"])
        else:
            raise ValueError(f"Unsupported type `{instance_type}` for label graph")

        label_map = {label_id: all_labels.get(IDMapper().backward(label_id)) for label_id in instance["nodes"]}
        for label in label_map.values():
            output.add_node(label)
        for edge in instance["edges"]:
            output.add_edge(label_map[edge[0]], label_map[edge[1]])

        return output


class LabelSchemaMapper:
    """This class maps a `LabelSchema` entity to a serialized dictionary, and vice versa."""

    @staticmethod
    def forward(
        instance: LabelSchemaEntity,
    ) -> dict:
        """Serializes to dict."""

        label_groups = [LabelGroupMapper().forward(group) for group in instance.get_groups(include_empty=True)]

        return {
            "label_tree": LabelGraphMapper().forward(instance.label_tree),
            "label_groups": label_groups,
            "all_labels": {
                IDMapper().forward(label.id_): LabelMapper().forward(label) for label in instance.get_labels(True)
            },
        }

    @staticmethod
    def backward(instance: dict) -> LabelSchemaEntity:
        """Deserializes from dict."""

        all_labels = {
            IDMapper().backward(id): LabelMapper().backward(label) for id, label in instance["all_labels"].items()
        }

        label_tree = LabelGraphMapper().backward(instance["label_tree"], all_labels)
        label_groups = [
            LabelGroupMapper().backward(label_group, all_labels) for label_group in instance["label_groups"]
        ]
        output = LabelSchemaEntity(
            label_tree=cast(LabelTree, label_tree),
            label_groups=label_groups,
        )
        return output


def label_schema_to_bytes(label_schema: LabelSchemaEntity) -> bytes:
    """Returns json-serialized LabelSchemaEntity as bytes."""

    serialized_label_schema = LabelSchemaMapper.forward(label_schema)
    return json.dumps(serialized_label_schema, indent=4).encode()



"""Images capturing module."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import os
import sys

import cv2

# Taken from here:
# https://github.com/openvinotoolkit/open_model_zoo/blob/develop/demos/common/python/images_capture.py


class InvalidInput(Exception):
    """Exception for wrong input format."""

    def __init__(self, message):
        super().__init__()
        self.message = message


class OpenError(Exception):
    """Exception for error opening reader."""

    def __init__(self, message):
        super().__init__()
        self.message = message


class ImagesCapture:
    """Images capturing base class."""

    def read(self):
        """Returns captured image."""
        raise NotImplementedError

    def fps(self):
        """Returns a frequency of getting images from source."""
        raise NotImplementedError

    def get_type(self):
        """Returns type of image capture."""
        raise NotImplementedError


class ImreadWrapper(ImagesCapture):
    """Class for reading an image from file."""

    def __init__(self, source, loop):
        self.loop = loop
        if not os.path.isfile(source):
            raise InvalidInput(f"Can't find the image by {source}")
        self.image = cv2.imread(source, cv2.IMREAD_COLOR)
        if self.image is None:
            raise OpenError(f"Can't open the image from {source}")
        self.can_read = True

    def read(self):
        """Returns captured image."""
        if self.loop:
            return copy.deepcopy(self.image)
        if self.can_read:
            self.can_read = False
            return copy.deepcopy(self.image)
        return None

    def fps(self):
        """Returns a frequency of getting images from source."""
        return 1.0

    def get_type(self):
        """Returns type of image capture."""
        return "IMAGE"


class DirReader(ImagesCapture):
    """Class for reading images from directory."""

    def __init__(self, source, loop):
        self.loop = loop
        self.dir = source
        if not os.path.isdir(self.dir):
            raise InvalidInput(f"Can't find the dir by {source}")
        self.names = sorted(os.listdir(self.dir))
        if not self.names:
            raise OpenError(f"The dir {source} is empty")
        self.file_id = 0
        for name in self.names:
            filename = os.path.join(self.dir, name)
            image = cv2.imread(filename, cv2.IMREAD_COLOR)
            if image is not None:
                return
        raise OpenError(f"Can't read the first image from {source}")

    def read(self):
        """Returns captured image."""
        while self.file_id < len(self.names):
            filename = os.path.join(self.dir, self.names[self.file_id])
            image = cv2.imread(filename, cv2.IMREAD_COLOR)
            self.file_id += 1
            if image is not None:
                return image
        if self.loop:
            self.file_id = 0
            while self.file_id < len(self.names):
                filename = os.path.join(self.dir, self.names[self.file_id])
                image = cv2.imread(filename, cv2.IMREAD_COLOR)
                self.file_id += 1
                if image is not None:
                    return image
        return None

    def fps(self):
        """Returns a frequency of getting images from source."""
        return 1.0

    def get_type(self):
        """Returns type of image capture."""
        return "DIR"


class VideoCapWrapper(ImagesCapture):
    """Class for capturing images from video."""

    def __init__(self, source, loop):
        self.loop = loop
        self.cap = cv2.VideoCapture()
        status = self.cap.open(source)
        if not status:
            raise InvalidInput(f"Can't open the video from {source}")

    def read(self):
        """Returns captured image."""
        status, image = self.cap.read()
        if not status:
            if not self.loop:
                return None
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
            status, image = self.cap.read()
            if not status:
                return None
        return image

    def fps(self):
        """Returns a frequency of getting images from source."""
        return self.cap.get(cv2.CAP_PROP_FPS)

    def get_type(self):
        """Returns type of image capture."""
        return "VIDEO"


class CameraCapWrapper(ImagesCapture):
    """Class for capturing images from camera."""

    def __init__(self, source, camera_resolution):
        self.cap = cv2.VideoCapture()
        try:
            status = self.cap.open(int(source))
            self.cap.set(cv2.CAP_PROP_BUFFERSIZE, 1)
            self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, camera_resolution[0])
            self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, camera_resolution[1])
            self.cap.set(cv2.CAP_PROP_FPS, 30)
            self.cap.set(cv2.CAP_PROP_AUTOFOCUS, 1)
            self.cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*"MJPG"))
            if not status:
                raise OpenError(f"Can't open the camera from {source}")
        except ValueError as ex:
            raise InvalidInput(f"Can't find the camera {source}") from ex

    def read(self):
        """Returns captured image."""
        status, image = self.cap.read()
        if not status:
            return None
        return image

    def fps(self):
        """Returns a frequency of getting images from source."""
        return self.cap.get(cv2.CAP_PROP_FPS)

    def get_type(self):
        """Returns type of image capture."""
        return "CAMERA"


def open_images_capture(source, loop, camera_resolution=(1280, 720)):
    """Opens images capture."""

    errors = {InvalidInput: [], OpenError: []}
    for reader in (ImreadWrapper, DirReader, VideoCapWrapper):
        try:
            return reader(source, loop)
        except (InvalidInput, OpenError) as ex:
            errors[type(ex)].append(ex.message)
    try:
        return CameraCapWrapper(source, camera_resolution)
    except (InvalidInput, OpenError) as ex:
        errors[type(ex)].append(ex.message)
    if not errors[OpenError]:
        print(*errors[InvalidInput], file=sys.stderr, sep="\n")
    else:
        print(*errors[OpenError], file=sys.stderr, sep="\n")
    sys.exit(1)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os.path as osp
import platform
import shutil
import sys
import tempfile
import warnings
from importlib import import_module
from typing import Any, Callable, Union

from mmcv.utils import Config, ConfigDict
from mmcv.utils.config import BASE_KEY, DEPRECATION_KEY
from mmcv.utils.misc import import_modules_from_strings
from mmcv.utils.path import check_file_exist

from otx.mpa.utils.logger import get_logger

logger = get_logger()


class MPAConfig(Config):
    @staticmethod
    def _file2dict(filename, use_predefined_variables=True):
        filename = osp.abspath(osp.expanduser(filename))
        check_file_exist(filename)
        fileExtname = osp.splitext(filename)[1]
        if fileExtname not in [".py", ".json", ".yaml", ".yml"]:
            raise IOError("Only py/yml/yaml/json type are supported now!")

        with tempfile.TemporaryDirectory() as temp_config_dir:
            temp_config_file = tempfile.NamedTemporaryFile(dir=temp_config_dir, suffix=fileExtname)
            if platform.system() == "Windows":
                temp_config_file.close()
            temp_config_name = osp.basename(temp_config_file.name)
            # Substitute predefined variables
            if use_predefined_variables:
                Config._substitute_predefined_vars(filename, temp_config_file.name)
            else:
                shutil.copyfile(filename, temp_config_file.name)
            # Substitute base variables from placeholders to strings
            base_var_dict = Config._pre_substitute_base_vars(temp_config_file.name, temp_config_file.name)
            if filename.endswith(".py"):
                temp_module_name = osp.splitext(temp_config_name)[0]
                sys.path.insert(0, temp_config_dir)
                Config._validate_py_syntax(filename)
                mod = import_module(temp_module_name)
                sys.path.pop(0)
                cfg_dict = {name: value for name, value in mod.__dict__.items() if not name.startswith("__")}
                # delete imported module
                del sys.modules[temp_module_name]
            elif filename.endswith((".yml", ".yaml", ".json")):
                import mmcv

                cfg_dict = mmcv.load(temp_config_file.name)
            # close temp file
            temp_config_file.close()

        # check deprecation information
        if DEPRECATION_KEY in cfg_dict:
            deprecation_info = cfg_dict.pop(DEPRECATION_KEY)
            warning_msg = f"The config file {filename} will be deprecated " "in the future."
            if "expected" in deprecation_info:
                warning_msg += f' Please use {deprecation_info["expected"]} ' "instead."
            if "reference" in deprecation_info:
                warning_msg += " More information can be found at " f'{deprecation_info["reference"]}'
            warnings.warn(warning_msg)

        cfg_text = filename + "\n"
        with open(filename, "r", encoding="utf-8") as f:
            # Setting encoding explicitly to resolve coding issue on windows
            cfg_text += f.read()

        if BASE_KEY in cfg_dict:
            cfg_dir = osp.dirname(filename)
            base_filename = cfg_dict.pop(BASE_KEY)
            base_filename = base_filename if isinstance(base_filename, list) else [base_filename]

            cfg_dict_list = list()
            cfg_text_list = list()
            for f in base_filename:
                _cfg_dict, _cfg_text = MPAConfig._file2dict(osp.join(cfg_dir, f))
                cfg_dict_list.append(_cfg_dict)
                cfg_text_list.append(_cfg_text)

            base_cfg_dict = dict()
            # for c in cfg_dict_list:
            #     duplicate_keys = base_cfg_dict.keys() & c.keys()
            #     if len(duplicate_keys) > 0:
            #         raise KeyError('Duplicate key is not allowed among bases. '
            #                        f'Duplicate keys: {duplicate_keys}')
            #     base_cfg_dict.update(c)
            for c in cfg_dict_list:
                if len(base_cfg_dict.keys() & c.keys()) > 0:
                    # raise KeyError(f'Duplicate key is not allowed among bases [{base_cfg_dict.keys() & c.keys()}]')
                    logger.warning(f"Duplicate key is detected among bases [{base_cfg_dict.keys() & c.keys()}]")
                    logger.debug(f"base = {base_cfg_dict}, cfg = {c}")
                    base_cfg_dict = Config._merge_a_into_b(base_cfg_dict, c)
                    logger.debug(f"merged dict = {base_cfg_dict}")
                else:
                    base_cfg_dict.update(c)

            # Subtitute base variables from strings to their actual values
            cfg_dict = Config._substitute_base_vars(cfg_dict, base_var_dict, base_cfg_dict)

            base_cfg_dict = Config._merge_a_into_b(cfg_dict, base_cfg_dict)
            cfg_dict = base_cfg_dict

            # merge cfg_text
            cfg_text_list.append(cfg_text)
            cfg_text = "\n".join(cfg_text_list)

        return cfg_dict, cfg_text

    @staticmethod
    def fromfile(filename, use_predefined_variables=True, import_custom_modules=True):
        cfg_dict, cfg_text = MPAConfig._file2dict(filename, use_predefined_variables)
        if import_custom_modules and cfg_dict.get("custom_imports", None):
            import_modules_from_strings(**cfg_dict["custom_imports"])
        return Config(cfg_dict, cfg_text=cfg_text, filename=filename)


def copy_config(cfg):
    if not isinstance(cfg, Config):
        ValueError(f"cannot copy this instance {type(cfg)}")
    # new_cfg = copy.deepcopy(cfg)
    # new_cfg._cfg_dict = copy.deepcopy(cfg._cfg_dict)
    # new_cfg.filename = cfg.filename
    import pickle

    data = pickle.dumps(cfg)
    return pickle.loads(data)


def update_or_add_custom_hook(cfg: Config, hook_cfg: ConfigDict):
    """Update hook cfg if same type is in custom_hook or append it"""
    custom_hooks = cfg.get("custom_hooks", [])
    custom_hooks_updated = False
    for custom_hook in custom_hooks:
        if custom_hook["type"] == hook_cfg["type"]:
            custom_hook.update(hook_cfg)
            custom_hooks_updated = True
            break
    if not custom_hooks_updated:
        custom_hooks.append(hook_cfg)
    cfg["custom_hooks"] = custom_hooks


def remove_custom_hook(cfg: Config, hook_type: str):
    """Remove hook cfg if hook_type is in custom_hook"""
    custom_hooks = cfg.get("custom_hooks", [])
    if len(custom_hooks) > 0:
        idx_to_del = None
        for i, custom_hook in enumerate(custom_hooks):
            if custom_hook["type"] == hook_type:
                idx_to_del = i
                break
        if idx_to_del is not None:
            del custom_hooks[idx_to_del]


def recursively_update_cfg(
    cfg: Union[Config, dict],
    criterion: Callable[[Any, Any], bool],
    update_dict: Any,
):
    for k, v in list(cfg.items()):
        if isinstance(v, dict):
            recursively_update_cfg(v, criterion, update_dict)
        if criterion(k, v):
            cfg.update(update_dict)


def add_custom_hook_if_not_exists(cfg: Config, hook_cfg: ConfigDict):
    custom_hooks = cfg.get("custom_hooks", [])
    found = False
    for hook in custom_hooks:
        if hook["type"] == hook_cfg["type"]:
            found = True
            break
    if not found:
        custom_hooks.append(hook_cfg)
        cfg["custom_hooks"] = custom_hooks



"""This module implements the TaskEnvironment entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import List, Optional, Type, TypeVar

from otx.api.configuration import ConfigurableParameters, cfg_helper
from otx.api.entities.label import LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.model_template import ModelTemplate

TypeVariable = TypeVar("TypeVariable", bound=ConfigurableParameters)


# pylint: disable=too-many-instance-attributes; Requires refactor
class TaskEnvironment:
    """Defines the machine learning environment the task runs in.

    Args:
    model_template (ModelTemplate): The model template used for this task
    model (Optional[ModelEntity]): Model to use; if not specified, the task must be either weight-less
        or use pre-trained or randomly initialised weights.
    hyper_parameters (ConfigurableParameters): Set of hyper parameters
    label_schema (LabelSchemaEntity): Label schema associated to this task
    """

    def __init__(
        self,
        model_template: ModelTemplate,
        model: Optional[ModelEntity],
        hyper_parameters: ConfigurableParameters,
        label_schema: LabelSchemaEntity,
    ):

        self.model_template = model_template
        self.model = model
        self.__hyper_parameters = hyper_parameters
        self.label_schema = label_schema

    def __repr__(self):
        """String representation of the TaskEnvironment object."""
        return (
            f"TaskEnvironment(model={self.model}, label_schema={self.label_schema}, "
            f"hyper_params={self.__hyper_parameters})"
        )

    def __eq__(self, other: object) -> bool:
        """Compares two TaskEnvironment objects.

        Args:
            other (TaskEnvironment): Object to compare with.

        Returns:
            bool: True if equal, False otherwise.
        """
        if isinstance(other, TaskEnvironment):
            return (
                self.model == other.model
                and self.label_schema == other.label_schema
                # TODO get_hyperparameters should return Union rather than TypeVariable
                and self.get_hyper_parameters(instance_of=None)  # type: ignore
                == other.get_hyper_parameters(instance_of=None)
            )
        return False

    def get_labels(self, include_empty: bool = False) -> List[LabelEntity]:
        """Return the labels in this task environment (based on the label schema).

        Args:
            include_empty (bool): Include the empty label if ``True``. Defaults to False.

        Returns:
            List[LabelEntity]: List of labels
        """
        return self.label_schema.get_labels(include_empty)

    def get_hyper_parameters(self, instance_of: Optional[Type[TypeVariable]] = None) -> TypeVariable:
        """Returns Configuration for the task, de-serialized as type specified in `instance_of`.

        If the type of the configurable parameters is unknown, a generic
        ConfigurableParameters object with all available parameters can be obtained
        by calling method with instance_of = None.

        Example:
            >>> self.get_hyper_parameters(instance_of=TorchSegmentationConfig)
            TorchSegmentationConfig()

        Args:
            instance_of (Optional[Type[TypeVariable]]): subtype of ModelConfig of the hyperparamters. Defaults to None.

        Returns:
            TypeVariable: ConfigurableParameters entity
        """
        if instance_of is None:
            # If the instance_of is None, the type variable is not defined so the
            # return type won't be deduced correctly
            return self.__hyper_parameters  # type: ignore

        # Otherwise, update the base config according to what is stored in the repo.
        base_config = instance_of(header=self.__hyper_parameters.header)

        cfg_helper.substitute_values(base_config, value_input=self.__hyper_parameters, allow_missing_values=True)

        return base_config

    def set_hyper_parameters(self, hyper_parameters: ConfigurableParameters):
        """Sets the hyper parameters for the task.

        Example:
            >>> self.set_hyper_parameters(hyper_parameters=TorchSegmentationParameters())
            None
        Args:
            hyper_parameters (ConfigurationParameter): ConfigurableParameters entity to assign to task
        """
        if not isinstance(hyper_parameters, ConfigurableParameters):
            raise ValueError(f"Unable to set hyper parameters, invalid input: {hyper_parameters}")
        self.__hyper_parameters = hyper_parameters

    def get_model_configuration(self) -> ModelConfiguration:
        """Get the configuration needed to use the current model.

        That is the current set of:
            * configurable parameters
            * labels
            * label schema

        Returns:
            ModelConfiguration: Model configuration
        """
        return ModelConfiguration(self.__hyper_parameters, self.label_schema)



"""Detection Dataset Adapter."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name, too-many-locals, no-member, too-many-nested-blocks
from typing import List

from datumaro.components.annotation import AnnotationType

from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.model_template import TaskType
from otx.core.data.adapter.base_dataset_adapter import BaseDatasetAdapter


class DetectionDatasetAdapter(BaseDatasetAdapter):
    """Detection adapter inherited from BaseDatasetAdapter.

    It converts DatumaroDataset --> DatasetEntity for object detection, and instance segmentation tasks
    """

    def get_otx_dataset(self) -> DatasetEntity:
        """Convert DatumaroDataset to DatasetEntity for Detection."""
        # Prepare label information
        label_information = self._prepare_label_information(self.dataset)
        self.label_entities = label_information["label_entities"]

        dataset_items: List[DatasetItemEntity] = []
        used_labels: List[int] = []
        for subset, subset_data in self.dataset.items():
            for _, datumaro_items in subset_data.subsets().items():
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    shapes = []
                    for ann in datumaro_item.annotations:
                        if (
                            self.task_type in (TaskType.INSTANCE_SEGMENTATION, TaskType.ROTATED_DETECTION)
                            and ann.type == AnnotationType.polygon
                        ):
                            if self._is_normal_polygon(ann):
                                shapes.append(self._get_polygon_entity(ann, image.width, image.height))
                        if self.task_type is TaskType.DETECTION and ann.type == AnnotationType.bbox:
                            if self._is_normal_bbox(ann.points[0], ann.points[1], ann.points[2], ann.points[3]):
                                shapes.append(self._get_normalized_bbox_entity(ann, image.width, image.height))

                        if ann.label not in used_labels:
                            used_labels.append(ann.label)
                    dataset_item = DatasetItemEntity(image, self._get_ann_scene_entity(shapes), subset=subset)
                    dataset_items.append(dataset_item)

        self.remove_unused_label_entities(used_labels)
        return DatasetEntity(items=dataset_items)



"""Convert a mask to a border image."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

from typing import List

import numpy as np
from skimage.measure import find_contours, label, regionprops


def mask_to_border(mask):
    """Make a border by using a binary mask.

    Args:
        mask (np.ndarray): Input binary mask

    Returns:
        np.ndarray: Border image.
    """
    h, w = mask.shape
    border = np.zeros((h, w))

    contours = find_contours(mask, 0.5)  # since the input range is [0, 1], the threshold is 0.5
    for contour in contours:
        for c in contour:
            x = int(c[0])
            y = int(c[1])
            border[x][y] = 1  # since the input is binary, the value is 1

    return border


def mask2bbox(mask):
    """Mask to bounding boxes.

    Args:
        mask (np.ndarray): Input binary mask

    Returns:
        List[int]: Bounding box coordinates
    """
    bboxes: List[int] = []

    mask = mask_to_border(mask)
    print(np.unique(mask))
    lbl = label(mask)
    props = regionprops(lbl)
    for prop in props:
        x1 = prop.bbox[1]
        y1 = prop.bbox[0]

        x2 = prop.bbox[3]
        y2 = prop.bbox[2]

        bboxes.append([x1, y1, x2, y2])

    return bboxes



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
"""
Dummy configurable parameter class to test configuration functionality
"""

import attr

from otx.api.configuration import (
    ConfigurableParameters,
    ModelLifecycle,
    Operator,
    Rule,
    UIRules,
)
from otx.api.configuration.elements import (
    ConfigurableEnum,
    ParameterGroup,
    add_parameter_group,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    float_selectable,
    selectable,
    string_attribute,
)
from otx.api.configuration.enums import AutoHPOState
from otx.api.configuration.ui_rules import Action


class SomeEnumSelectable(ConfigurableEnum):
    """Test Enum selectable"""

    TEST_NAME1 = "test_name_1"
    TEST_2 = "test_2_test"
    BOGUS_NAME = "bogus"
    OPTION_C = "option_c"


@attr.s
class DatasetManagerConfig(ConfigurableParameters):
    """Dummy configurable parameters class"""

    # type: ignore
    # This class is used for testing purposes only, so mypy should ignore it

    # Component and header are required, description is optional.
    header = string_attribute("Dataset Manager configuration -- TEST ONLY")
    description = string_attribute("Configurable parameters for the DatasetManager -- TEST ONLY")

    # Add some parameters
    number_of_samples_for_auto_train = configurable_integer(
        default_value=5,
        min_value=1,
        max_value=1000,
        header="Samples required for new training round",
    )
    label_constraints = configurable_boolean(default_value=True, header="Apply label constraints")

    @attr.s
    class _NestedParameterGroup(ParameterGroup):
        # A nested group of parameters
        # header is a required attribute, all parameter groups should define one.
        header = string_attribute("Test group of parameter groups")

        @attr.s
        class _SubgroupOne(ParameterGroup):
            # Subgroup one of the nested group, with a couple of parameters
            header = string_attribute("Parameter group one")

            __ui_rules = UIRules(
                rules=[
                    Rule(
                        parameter=["nested_parameter_group", "show_subgroup_one"],
                        operator=Operator.EQUAL_TO,
                        value=False,
                    )
                ],
                action=Action.HIDE,
            )

            bogus_parameter_one = configurable_float(
                default_value=42,
                ui_rules=__ui_rules,
                header="Bogus parameter to test nested parameter groups",
            )
            bogus_parameter_two = configurable_float(
                default_value=42,
                ui_rules=__ui_rules,
                header="Bogus parameter to test nested parameter groups",
            )

        subgroup_one = add_parameter_group(_SubgroupOne)

        show_subgroup_one = configurable_boolean(default_value=True, header="Show the parameters in subgroup one?")

        @attr.s
        class _SubgroupTwo(ParameterGroup):
            # Subgroup two of the nested group, with a couple of parameters
            header = string_attribute("Parameter group two")
            bogus_parameter_three = configurable_float(
                default_value=42,
                header="Bogus parameter to test nested parameter groups",
            )

            bogus_parameter_four = configurable_float(
                default_value=42,
                header="Bogus parameter to test nested parameter groups",
            )

        subgroup_two = add_parameter_group(_SubgroupTwo)

    @attr.s
    class _SubsetParameters(ParameterGroup):
        # Parameters governing sample distribution over subsets
        header = string_attribute("Subset parameters")
        description = string_attribute("Parameters for the different subsets")

        # Add a parameter group 'Subset parameters'
        auto_subset_fractions = configurable_boolean(
            default_value=True,
            description="Test",
            header="Automatically determine subset proportions",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        __ui_rules = UIRules(
            rules=[
                Rule(
                    parameter="auto_subset_fractions",
                    value=False,
                    operator=Operator.EQUAL_TO,
                )
            ],
            action=Action.SHOW,
        )

        train_proportion = configurable_float(
            default_value=0.75,
            min_value=0.0,
            max_value=1.0,
            header="Training set proportion",
            ui_rules=__ui_rules,
            affects_outcome_of=ModelLifecycle.TRAINING,
            auto_hpo_state=AutoHPOState.POSSIBLE,
        )

        validation_proportion = configurable_float(
            default_value=0.1,
            min_value=0.0,
            max_value=1.0,
            header="Validation set proportion",
            ui_rules=__ui_rules,
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        test_proportion = configurable_float(
            default_value=0.15,
            min_value=0.0,
            max_value=1.0,
            header="Test set proportion",
            ui_rules=__ui_rules,
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    # Add a selectable and float selectable parameter
    dummy_float_selectable = float_selectable(
        options=[1.0, 2.0, 3.0, 4.0],
        default_value=2.0,
        header="Test float selectable",
        auto_hpo_state=AutoHPOState.POSSIBLE,
    )

    dummy_selectable = selectable(
        default_value=SomeEnumSelectable.BOGUS_NAME,
        header="Test",
        affects_outcome_of=ModelLifecycle.INFERENCE,
    )

    # Finally, add the nested parameter group and subset parameter groups to the config
    # NOTE! group initialization should use a factory to avoid passing mutable default arguments. This is why the
    # add_parameter_group function is needed.
    nested_parameter_group = add_parameter_group(_NestedParameterGroup)
    subset_parameters = add_parameter_group(_SubsetParameters)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmseg.models.builder import HEADS
from mmseg.models.decode_heads.fcn_head import FCNHead

from .aggregator_mixin import AggregatorMixin
from .mix_loss_mixin import MixLossMixin
from .pixel_weights_mixin import PixelWeightsMixin2
from .segment_out_norm_mixin import SegmentOutNormMixin


@HEADS.register_module()
class CustomFCNHead(SegmentOutNormMixin, AggregatorMixin, MixLossMixin, PixelWeightsMixin2, FCNHead):
    """Custom Fully Convolution Networks for Semantic Segmentation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # get rid of last activation of convs module
        if self.act_cfg:
            self.convs[-1].with_activation = False
            delattr(self.convs[-1], "activate")

        if kwargs.get("init_cfg", {}):
            self.init_weights()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn
from mmcv.runner import HOOKS, Hook

from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class NoBiasDecayHook(Hook):
    """Hook for No Bias Decay Method (Bag of Tricks for Image Classification)

    This hook divides model's weight & bias to 3 parameter groups
    [weight with decay, weight without decay, bias without decay]
    """

    def before_train_epoch(self, runner):
        weight_decay, bias_no_decay, weight_no_decay = [], [], []
        for m in runner.model.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                weight_decay.append(m.weight)
                if m.bias is not None:
                    bias_no_decay.append(m.bias)
            elif hasattr(m, "weight") or hasattr(m, "bias"):
                if hasattr(m, "weight"):
                    weight_no_decay.append(m.weight)
                if hasattr(m, "bias"):
                    bias_no_decay.append(m.bias)
            elif len(list(m.children())) == 0:
                for p in m.parameters():
                    weight_decay.append(p)

        weight_decay_group = runner.optimizer.param_groups[0].copy()
        weight_decay_group["params"] = weight_decay

        bias_group = runner.optimizer.param_groups[0].copy()
        bias_group["params"] = bias_no_decay
        bias_group["weight_decay"] = 0.0
        bias_group["lr"] *= 2

        weight_no_decay_group = runner.optimizer.param_groups[0].copy()
        weight_no_decay_group["params"] = weight_no_decay
        weight_no_decay_group["weight_decay"] = 0.0

        param_groups = [weight_decay_group, bias_group, weight_no_decay_group]
        runner.optimizer.param_groups = param_groups

    def after_train_epoch(self, runner):
        params = []
        for m in runner.model.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                params.append(m.weight)
                if m.bias is not None:
                    params.append(m.bias)
            elif hasattr(m, "weight") or hasattr(m, "bias"):
                if hasattr(m, "weight"):
                    params.append(m.weight)
                if hasattr(m, "bias"):
                    params.append(m.bias)
            elif len(list(m.children())) == 0:
                for p in m.parameters():
                    params.append(p)

        param_groups = runner.optimizer.param_groups[0].copy()
        param_groups["params"] = params
        runner.optimizer.param_groups = [param_groups]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import tempfile

import numpy as np
import torch
from mmcv.utils import Config

from otx.mpa.deploy.apis import NaiveExporter
from otx.mpa.deploy.utils import is_mmdeploy_enabled
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.mpa.deploy.test_helpers import create_config, create_model


class TestNaiveExporter:
    @e2e_pytest_unit
    def test_sub_component(self):
        model = create_model("mmcls")

        with tempfile.TemporaryDirectory() as tempdir:
            onnx_path = NaiveExporter.torch2onnx(
                tempdir,
                model,
                {"img": [torch.zeros((1, 50, 50, 3))], "img_metas": []},
            )
            assert os.path.exists(onnx_path)

            openvino_paths = NaiveExporter.onnx2openvino(
                tempdir,
                onnx_path,
            )
            for openvino_path in openvino_paths:
                assert os.path.exists(openvino_path)

    @e2e_pytest_unit
    def test_export2openvino(self):
        from otx.algorithms.classification.adapters.mmcls.utils.builder import (
            build_classifier,
        )

        config = create_config()
        create_model("mmcls")

        with tempfile.TemporaryDirectory() as tempdir:
            NaiveExporter.export2openvino(
                tempdir,
                build_classifier,
                config,
                {"img": [torch.zeros((50, 50, 3))], "img_metas": []},
            )
            assert [f for f in os.listdir(tempdir) if f.endswith(".xml")]
            assert [f for f in os.listdir(tempdir) if f.endswith(".bin")]


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER, mark

    from otx.mpa.deploy.apis import MMdeployExporter

    class TestMMdeployExporter:
        @e2e_pytest_unit
        def test_sub_component(self):
            config = create_config()
            deploy_config = Config(
                {
                    "ir_config": {
                        "type": "onnx",
                        "input_names": ["input"],
                        "output_names": ["output"],
                    },
                    "codebase_config": {
                        "type": "mmcls",
                        "task": "Classification",
                    },
                    "backend_config": {
                        "type": "openvino",
                        "model_inputs": [
                            {
                                "opt_shapes": {
                                    "input": [1, 3, 50, 50],
                                }
                            }
                        ],
                    },
                }
            )
            create_model("mmcls")

            with tempfile.TemporaryDirectory() as tempdir:
                onnx_path = MMdeployExporter.torch2onnx(
                    tempdir,
                    np.zeros((50, 50, 3), dtype=np.float32),
                    config,
                    deploy_config,
                )
                assert isinstance(onnx_path, str)
                assert os.path.exists(onnx_path)

                openvino_paths = MMdeployExporter.onnx2openvino(
                    tempdir,
                    onnx_path,
                    deploy_config,
                )
                for openvino_path in openvino_paths:
                    assert os.path.exists(openvino_path)

        @e2e_pytest_unit
        def test_export2openvino(self):
            from otx.algorithms.classification.adapters.mmcls.utils.builder import (
                build_classifier,
            )

            config = create_config()
            deploy_config = Config(
                {
                    "ir_config": {
                        "type": "onnx",
                        "input_names": ["input"],
                        "output_names": ["output"],
                    },
                    "codebase_config": {
                        "type": "mmcls",
                        "task": "Classification",
                    },
                    "backend_config": {
                        "type": "openvino",
                        "model_inputs": [
                            {
                                "opt_shapes": {
                                    "input": [1, 3, 50, 50],
                                }
                            }
                        ],
                    },
                }
            )
            create_model("mmcls")

            with tempfile.TemporaryDirectory() as tempdir:
                MMdeployExporter.export2openvino(
                    tempdir,
                    build_classifier,
                    config,
                    deploy_config,
                )
                assert [f for f in os.listdir(tempdir) if f.endswith(".xml")]
                assert [f for f in os.listdir(tempdir) if f.endswith(".bin")]

        @e2e_pytest_unit
        def test_partition(self):
            from otx.algorithms.classification.adapters.mmcls.utils.builder import (
                build_classifier,
            )

            config = create_config()
            deploy_config = Config(
                {
                    "ir_config": {
                        "type": "onnx",
                        "input_names": ["input"],
                        "output_names": ["output"],
                    },
                    "codebase_config": {
                        "type": "mmcls",
                        "task": "Classification",
                    },
                    "backend_config": {
                        "type": "openvino",
                        "model_inputs": [
                            {
                                "opt_shapes": {
                                    "input": [1, 3, 50, 50],
                                }
                            }
                        ],
                    },
                    "partition_config": {
                        "apply_marks": True,
                        "partition_cfg": [
                            {
                                "save_file": "partition.onnx",
                                "start": ["test:input"],
                                "end": ["test:output"],
                                "output_names": ["output"],
                                "mo_options": {
                                    "_delete_": True,
                                    "args": {},
                                    "flags": [],
                                },
                            }
                        ],
                    },
                }
            )
            create_model("mmcls")

            @FUNCTION_REWRITER.register_rewriter("tests.unit.mpa.deploy.test_helpers.MockModel.forward")
            @mark("test", inputs=["input"], outputs=["output"])
            def forward(ctx, self, *args, **kwargs):
                return ctx.origin_func(self, *args, **kwargs)

            with tempfile.TemporaryDirectory() as tempdir:
                MMdeployExporter.export2openvino(
                    tempdir,
                    build_classifier,
                    config,
                    deploy_config,
                )
                files = os.listdir(tempdir)
                assert "model.onnx" in files
                assert "model.xml" in files
                assert "model.bin" in files
                assert "partition.onnx" in files
                assert "partition.xml" in files
                assert "partition.bin" in files



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass

import torch

from .builder import OPS
from .op import Attribute, Operation
from .type_conversions import _ov_to_torch


@dataclass
class RangeV4Attribute(Attribute):
    output_type: str


@OPS.register()
class RangeV4(Operation[RangeV4Attribute]):
    TYPE = "Range"
    VERSION = 4
    ATTRIBUTE_FACTORY = RangeV4Attribute

    def forward(self, start, stop, step):
        dtype = _ov_to_torch[self.attrs.output_type]
        return torch.arange(
            start=start,
            end=stop,
            step=step,
            dtype=dtype,
            device=start.device,
            requires_grad=False,
        )



"""Unit Test for otx.algorithms.action.adapters.mmaction.utils.config_utils."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, List

import numpy as np
import torch
from mmaction.models import Recognizer3D
from mmcv.runner import BaseModule
from mmcv.utils import Config
from torch import nn

from otx.algorithms.action.adapters.mmaction.models.detectors.fast_rcnn import (
    AVAFastRCNN,
)
from otx.algorithms.action.adapters.mmaction.utils.export_utils import (
    _convert_sync_batch_to_normal_batch,
    export_model,
    get_frame_inds,
    onnx2openvino,
    preprocess,
    pytorch2onnx,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockRecognizer3D(Recognizer3D, BaseModule):
    """Mock class for Recognizer3D."""

    def __init__(self) -> None:
        super(BaseModule, self).__init__()

    def forward_dummy(self, inputs: Any) -> str:
        return "Forward function is replaced!"

    @staticmethod
    def cpu() -> Recognizer3D:
        return MockRecognizer3D()


class MockAVAFastRCNN(AVAFastRCNN):
    """Mock class for AVAFastRCNN."""

    def __init__(self) -> None:
        super(BaseModule, self).__init__()

    def add_detector(self) -> None:
        pass

    def patch_pools(self) -> None:
        pass

    def forward_infer(self, inputs: Any, img_metas: Any) -> str:
        return "Forward function is replaced!"

    @staticmethod
    def cpu() -> AVAFastRCNN:
        return MockAVAFastRCNN()


def _mock_sync_batchnorm(inputs):
    """Mock function for _sync_batch_to_normal_batch function.

    It returns its inputs
    """

    return inputs


@e2e_pytest_unit
def test_convert_sync_batch_to_normal_batch() -> None:
    """Test _convert_sync_batch_to_normal_batch function.

    <Steps>
        1. Create sample module, which has some Conv3D, SyncBatchNorm, BatchNorm3d ops
        2. Run _convert_sync_batch_to_normal_batch function to sample module
        3. Check SyncBatchNorm is changed into BatchNorm3d
        4. Check the other ops don't affect by this function
    """

    sample_module = nn.Sequential(
        nn.Conv3d(100, 100, 3), nn.SyncBatchNorm(100), nn.Conv3d(100, 100, 3), nn.BatchNorm3d(100)
    )
    output_module = _convert_sync_batch_to_normal_batch(sample_module)
    assert isinstance(output_module[0], nn.Conv3d)
    assert isinstance(output_module[1], nn.BatchNorm3d)
    assert isinstance(output_module[2], nn.Conv3d)
    assert isinstance(output_module[3], nn.BatchNorm3d)


@e2e_pytest_unit
def test_preprocess(mocker) -> None:
    """Test preprocess function.

    <Steps>
        1. Run preprocess function
        2. Check output imgs has proper shape: [1, 3, 3, 256, 256]
        2. Check output meta contains proper information
    """

    mocker.patch(
        "otx.algorithms.action.adapters.mmaction.utils.export_utils.get_frame_inds", return_value=np.array([2, 4, 6])
    )
    imgs, meta = preprocess(clip_len=8, width=256, height=256)
    assert list(imgs.shape) == [1, 3, 3, 256, 256]
    assert torch.all(meta[0][0]["img_shape"] == torch.Tensor([256.0, 256.0]))
    assert meta[0][0]["ori_shape"] == (5, 10)
    assert np.all(meta[0][0]["scale_factor"] == np.array([256 / 10, 256 / 5, 256 / 10, 256 / 5]))


@e2e_pytest_unit
def test_get_frame_inds() -> None:
    """Test get_frame_inds function.

    <Steps>
        1. Run get_frame_inds function for long enough inputs
        2. Check output indices
        3. Run get_frame_inds function for short inputs
        4. Check output indices
    """

    inputs = np.ndarray([1, 3, 50, 256, 256])
    outputs = get_frame_inds(inputs, 8, 2)
    assert np.all(outputs == np.array([17, 19, 21, 23, 25, 27, 29, 31]))

    inputs = np.ndarray([1, 3, 10, 256, 256])
    outputs = get_frame_inds(inputs, 8, 2)
    assert np.all(outputs == np.array([0, 2, 4, 6, 8, 9, 9, 9]))


@e2e_pytest_unit
def test_pytorch2onnx(mocker) -> None:
    """Test pytorch2onnx function.

    It checks whether model's forward function is replaced with proper alternatives
    <Steps>
        1. Run pytorch2onnx function for action detection model
        2. Check model's forward function is replaced with forward_infer function
        3. Run pytorch2onnx function for action classification model
        4. Check model's forward function is replaced with forward_dummy function
    """

    sample_tensor = torch.randn(1, 3, 32, 256, 256)
    sample_meta = [[{"meta": None}]]

    mocker.patch(
        "otx.algorithms.action.adapters.mmaction.utils.export_utils._convert_sync_batch_to_normal_batch",
        side_effect=_mock_sync_batchnorm,
    )
    mocker.patch(
        "otx.algorithms.action.adapters.mmaction.utils.export_utils.preprocess",
        return_value=(sample_tensor, sample_meta),
    )
    mocker.patch("otx.algorithms.action.adapters.mmaction.utils.export_utils.torch.onnx.export", return_value=True)

    model = MockAVAFastRCNN()
    pytorch2onnx(model, [1, 3, 32, 256, 256])
    assert model(sample_tensor) == "Forward function is replaced!"

    model = MockRecognizer3D()
    pytorch2onnx(model, [1, 1, 3, 8, 224, 224])
    assert model(sample_tensor) == "Forward function is replaced!"


@e2e_pytest_unit
def test_onnx2openvino(mocker) -> None:
    """Test onnx2openvino function.

    It checks the final openvino running command
    <Steps>
        1. Run onnx2openvino function for MockCfg
        2. Check openvino model conversion command
    """

    MockCfg = Config(
        {
            "model": {"pretrained": "./weights.pth"},
            "data": {
                "test": {
                    "test_mode": False,
                    "pipeline": [
                        {"type": "Empty"},
                        {"type": "Normalize", "mean": [0.0, 0.0, 0.0], "std": [1.0, 1.0, 1.0], "to_bgr": True},
                    ],
                }
            },
        }
    )
    # TODO Check MockCfg

    MockModel = Config({"graph": {"output": [{"name": "1"}, {"name": "2"}, {"name": "3"}], "node": [{"output": "4"}]}})

    def _MockRun(command_line: List[str], **kwargs) -> None:
        """Mock function for subprocess.run function.

        It saves input command line into MockCfg so that we can check the command line
        """

        MockCfg.command_line = command_line

    mocker.patch("otx.algorithms.action.adapters.mmaction.utils.export_utils.onnx.load", return_value=MockModel)
    mocker.patch("otx.algorithms.action.adapters.mmaction.utils.export_utils.onnx.save", return_value=True)
    mocker.patch("otx.algorithms.action.adapters.mmaction.utils.export_utils.run", side_effect=_MockRun)

    onnx2openvino(
        cfg=MockCfg,
        onnx_model_path="temp.onnx",
        output_dir_path="./temp_dir/",
        layout="layout_sample",
        input_shape=[1, 3, 8, 224, 224],
        pruning_transformation=True,
    )

    assert MockCfg.model.pretrained is None
    assert MockCfg.data.test.test_mode is True
    assert MockCfg.command_line[0] in ["mo", "mo.py"]
    assert MockCfg.command_line[1].split("=")[-1] == "temp.onnx"
    assert MockCfg.command_line[2].split("=")[-1] == "[0.0, 0.0, 0.0]"
    assert MockCfg.command_line[3].split("=")[-1] == "[1.0, 1.0, 1.0]"
    assert MockCfg.command_line[4].split("=")[-1] == "./temp_dir/"
    assert MockCfg.command_line[7].split("=")[-1] == "layout_sample"
    assert MockCfg.command_line[8].split("=")[-1] == "[1, 3, 8, 224, 224]"
    assert MockCfg.command_line[9].split("--")[-1] == "reverse_input_channels"
    assert MockCfg.command_line[-1] == "Pruning"


@e2e_pytest_unit
def test_export_model(mocker) -> None:
    """Test export_model function.

    <Steps>
        1. Run export_model function for action classification model
        2. Check input_shape, and layout using MockCfg
        3. Run export_model funciton for action detection model
        4. Check input_shape, and layour using MockCfg
    """

    MockCfg = Config()

    def _MockPytorch2ONNX(*args, input_shape: List[int], **kwargs) -> None:
        """Mock function for pytorch2onnx.

        It saves input_shape to MockConfig
        """

        MockCfg.input_shape = input_shape

    def _MockONNX2OpenVINO(config, onnx_model_path, output_dir_path, layout, input_shape, **kwargs) -> None:
        """Mock function for onnx2pytorch.

        It saves layout to MockCfg
        """

        MockCfg.layout = layout

    mocker.patch(
        "otx.algorithms.action.adapters.mmaction.utils.export_utils.pytorch2onnx", side_effect=_MockPytorch2ONNX
    )
    mocker.patch(
        "otx.algorithms.action.adapters.mmaction.utils.export_utils.onnx2openvino", side_effect=_MockONNX2OpenVINO
    )

    export_model(MockRecognizer3D(), MockCfg)
    assert MockCfg.input_shape == [1, 1, 3, 8, 224, 224]
    assert MockCfg.layout == "??c???"

    export_model(MockAVAFastRCNN(), MockCfg)
    assert MockCfg.input_shape == [1, 3, 32, 256, 256]
    assert MockCfg.layout == "bctwh"



"""Unit Test for otx.algorithms.action.utils.convert_public_data_to_cvat."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import numpy as np
import pytest

from otx.algorithms.action.utils.convert_public_data_to_cvat import (
    convert_action_cls_dataset_to_datumaro,
    convert_ava_dataset_to_datumaro,
    generate_default_cvat_xml_fields,
    main,
    read_ava_csv,
    rename_and_copy,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockPath:
    """Mock class for pathlib.Path"""

    def __init__(self, path):
        self.path = path

    def mkdir(self, *args, **kwargs):
        return self.path


class MockFileObject:
    """Mock class for python default File object."""

    def readlines(self):
        lines = [
            "# Some comment\n",
            "video_dir frame_len class_idx\n",
            "video_dir frame_len class_idx\n",
            "video_dir frame_len class_idx\n",
            "video_dir frame_len class_idx\n",
            "video_dir frame_len class_idx\n",
        ]
        return lines

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        return None


class MockCsvFile:
    """Mock class for python default File object."""

    def __init__(self, *args, **kwargs):
        self.lines = [
            "0,0,0,0,1,1,0",
            "0,1,0,0,1,1,0",
            "0,2,0,0,1,1,0",
            "0,3,0,0,1,1,0",
            "0,4,0,0,1,1,0",
        ]

    def __enter__(self):
        return self.lines

    def __exit__(self, *args, **kwargs):
        return None


class MockElementTree:
    """Mock class for lxml.etree.ElementTree."""

    def write(self, *args, **kwargs):
        pass


@e2e_pytest_unit
def test_generate_default_cvat_xml_fields(mocker) -> None:
    """Test generate_default_cvat_xml_fields function."""
    video_path = "dummy_path"
    frame_list = ["dummy_frame0", "dummy_frame1", "dummy_frame2", "dummy_frame3"]

    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.cv2.imread", return_value=np.ndarray((256, 256, 3))
    )

    output = generate_default_cvat_xml_fields(1, video_path, frame_list)
    assert len(output[0].getchildren()) == 2
    assert output[1] == (256, 256, 3)
    assert len(output[2].getchildren()) == 0


@e2e_pytest_unit
def test_convert_action_cls_dataset_to_datumaro(mocker) -> None:
    """Test convert_jester_dataset_to_datumaro function."""

    src_path = "dummy_src_path"
    dst_path = "dummy_dst_path"
    ann_file = "dummy_ann_file"

    mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.open", return_value=MockFileObject())
    mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.pathlib.Path.mkdir", return_value=True)
    # mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.os.makedirs", return_value=True)
    mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.shutil.copy", return_value=True)
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.generate_default_cvat_xml_fields",
        return_value=([], (256, 256, 3), []),
    )
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.os.listdir", return_value=(["frame0", "frame1"])
    )
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.etree.ElementTree", return_value=MockElementTree()
    )
    convert_action_cls_dataset_to_datumaro(src_path, dst_path, ann_file)


@e2e_pytest_unit
def test_convert_ava_dataset_to_datumaro(mocker) -> None:
    """Test convert_ava_dataset_to_datumaro function."""

    src_path = "dummy_src_path"
    dst_path = "dummy_dst_path"
    ann_file = "dummy_ann_file"

    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.read_ava_csv",
        return_value={"video_0": {"frame_idx": [[0, 0, 1, 1, "action"]]}},
    )
    mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.os.listdir", return_value=["video_0"])
    mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.shutil.copytree", return_value=True)
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.generate_default_cvat_xml_fields",
        return_value=([], (256, 256, 3), []),
    )
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.etree.ElementTree", return_value=MockElementTree()
    )
    convert_ava_dataset_to_datumaro(src_path, dst_path, ann_file)


@e2e_pytest_unit
def test_rename_and_copy(mocker) -> None:
    mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.shutil.copy2", return_value=True)
    frame_name = "root/vid/frame_1.png"
    rename_and_copy(frame_name, frame_name)
    frame_name = "root/vid/1.png"
    rename_and_copy(frame_name, frame_name)


@e2e_pytest_unit
def test_read_ava_csv(mocker) -> None:
    mocker.patch("otx.algorithms.action.utils.convert_public_data_to_cvat.open", return_value=MockCsvFile())
    annot_info = read_ava_csv("dummy_path")
    assert len(annot_info) == 1
    assert len(annot_info["0"]) == 5
    assert annot_info["0"][0] == [["0", "0", "1", "1", "0"]]


@e2e_pytest_unit
@pytest.mark.parametrize("task", ["action_classification", "action_detection", "pose_estimation"])
def test_main(task, mocker) -> None:
    """Test main function."""
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.parse_args", return_value=mocker.MagicMock(task=task)
    )
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.convert_action_cls_dataset_to_datumaro",
        return_value=True,
    )
    mocker.patch(
        "otx.algorithms.action.utils.convert_public_data_to_cvat.convert_ava_dataset_to_datumaro", return_value=True
    )
    main()



_base_ = ["./train.py", "../_base_/data/coco_inst_seg.py", "../_base_/models/detectors/detector.py"]

task = "instance-segmentation"

evaluation = dict(
    interval=1, metric="mAP", save_best="mAP", iou_thr=[0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]
)

task_adapt = dict(
    type="mpa",
    op="REPLACE",
    efficient_mode=False,
)

runner = dict(max_epochs=300)

optimizer_config = dict(_delete_=True, grad_clip=dict(max_norm=35, norm_type=2))

ignore = True
adaptive_validation_interval = dict(
    max_interval=5,
    enable_adaptive_interval_hook=True,
    enable_eval_before_run=True,
)
adaptive_ema = dict(epoch_momentum=0.4)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.distributed as dist
import torch.utils.data.distributed


def get_dist_info():
    try:
        # data distributed parallel
        return dist.get_rank(), dist.get_world_size(), True
    except Exception:
        return 0, 1, False



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field

import torch
from torch.nn import functional as F

from .builder import OPS
from .op import Attribute, Operation
from .poolings import AvgPoolV1


@dataclass
class BatchNormalizationV0Attribute(Attribute):
    epsilon: float
    max_init_iter: int = field(default=2)


@OPS.register()
class BatchNormalizationV0(Operation[BatchNormalizationV0Attribute]):
    TYPE = "BatchNormInference"
    VERSION = 0
    ATTRIBUTE_FACTORY = BatchNormalizationV0Attribute

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_buffer("_num_init_iter", torch.tensor(0))

    def forward(self, input, gamma, beta, mean, variance):

        output = F.batch_norm(
            input=input,
            running_mean=mean,
            running_var=variance,
            weight=gamma,
            bias=beta,
            training=self.training,
            momentum=0.1,
            eps=self.attrs.epsilon,
        )

        if self.training and self._num_init_iter < self.attrs.max_init_iter:
            n_dims = input.dim() - 2
            gamma = gamma.unsqueeze(0)
            beta = beta.unsqueeze(0)
            for _ in range(n_dims):
                gamma = gamma.unsqueeze(-1)
                beta = beta.unsqueeze(-1)
            output = input * gamma + beta
            self._num_init_iter += 1
            if self._num_init_iter >= self.attrs.max_init_iter:
                # Adapt weight & bias using the first batch statistics
                # to undo normalization approximately
                gamma.data = gamma.data * mean
                beta.data = beta.data + (mean / (variance + self.attrs.epsilon))

        return output


@dataclass
class LocalResponseNormalizationV0Attribute(Attribute):
    alpha: float
    beta: float
    bias: float
    size: int


@OPS.register()
class LocalResponseNormalizationV0(Operation[LocalResponseNormalizationV0Attribute]):
    TYPE = "LRN"
    VERSION = 0
    ATTRIBUTE_FACTORY = LocalResponseNormalizationV0Attribute

    def forward(self, input, axes):
        dim = input.dim()

        axes = axes.detach().cpu().tolist()
        assert all(ax >= 1 for ax in axes)

        axes = [ax - 1 for ax in axes]
        kernel = [1 for _ in range(dim - 1)]
        stride = [1 for _ in range(dim - 1)]
        pads_begin = [0 for _ in range(dim - 1)]
        pads_end = [0 for _ in range(dim - 1)]
        for ax in axes:
            kernel[ax] = self.attrs.size
            pads_begin[ax] = self.attrs.size // 2
            pads_end[ax] = (self.attrs.size - 1) // 2

        avg_attrs = {
            "auto_pad": "explicit",
            "strides": stride,
            "kernel": kernel,
            "pads_begin": pads_begin,
            "pads_end": pads_end,
            "exclude-pad": True,
            "shape": self.shape,
        }
        avg_pool = AvgPoolV1("temp", **avg_attrs)

        div = input.mul(input).unsqueeze(1)
        div = avg_pool(div)
        div = div.squeeze(1)
        div = div.mul(self.attrs.alpha).add(self.attrs.bias).pow(self.attrs.beta)
        output = input / div
        return output


@dataclass
class NormalizeL2V0Attribute(Attribute):
    eps: float
    eps_mode: str

    def __post_init__(self):
        super().__post_init__()
        valid_eps_mode = ["add", "max"]
        if self.eps_mode not in valid_eps_mode:
            raise ValueError(f"Invalid eps_mode {self.eps_mode}. " f"It must be one of {valid_eps_mode}.")


@OPS.register()
class NormalizeL2V0(Operation[NormalizeL2V0Attribute]):
    TYPE = "NormalizeL2"
    VERSION = 0
    ATTRIBUTE_FACTORY = NormalizeL2V0Attribute

    def forward(self, input, axes):
        eps = self.attrs.eps
        eps_mode = self.attrs.eps_mode

        if isinstance(axes, torch.Tensor):
            axes = axes.detach().cpu().tolist()
        if not isinstance(axes, (list, tuple)):
            axes = [axes]

        # normalization layer convert to FP32 in FP16 training
        input_float = input.float()
        if axes:
            norm = input_float.pow(2).sum(axes, keepdim=True)
        else:
            norm = input_float

        if eps_mode == "add":
            norm = norm + eps
        elif eps_mode == "max":
            norm = torch.clamp(norm, max=eps)

        return (input_float / norm.sqrt()).type_as(input)


@dataclass
class MVNV6Attribute(Attribute):
    normalize_variance: bool
    eps: float
    eps_mode: str

    def __post_init__(self):
        super().__post_init__()
        valid_eps_mode = ["INSIDE_SQRT", "OUTSIDE_SQRT"]
        if self.eps_mode not in valid_eps_mode:
            raise ValueError(f"Invalid eps_mode {self.eps_mode}. " f"It must be one of {valid_eps_mode}.")


@OPS.register()
class MVNV6(Operation[MVNV6Attribute]):
    TYPE = "MVN"
    VERSION = 6
    ATTRIBUTE_FACTORY = MVNV6Attribute

    def forward(self, input, axes):
        output = input - input.mean(axes.tolist(), keepdim=True)
        if self.attrs.normalize_variance:
            eps_mode = self.attrs.eps_mode
            eps = self.attrs.eps
            var = torch.square(output).mean(axes.tolist(), keepdim=True)
            if eps_mode == "INSIDE_SQRT":
                output = output / torch.sqrt(var + eps)
            elif eps_mode == "OUTSIDE_SQRT":
                output = output / (torch.sqrt(var) + eps)
        return output



"""OTX Adapters - mmcv.utils."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from ._builder_build_data_parallel import build_data_parallel
from ._config_utils_get_configs_by_keys import get_configs_by_keys
from ._config_utils_get_configs_by_pairs import get_configs_by_pairs
from .builder import build_dataloader, build_dataset
from .config_utils import (
    align_data_config_with_recipe,
    config_from_string,
    get_data_cfg,
    get_dataset_configs,
    get_meta_keys,
    is_epoch_based_runner,
    patch_color_conversion,
    patch_data_pipeline,
    patch_default_config,
    patch_runner,
    prepare_for_testing,
    prepare_work_dir,
    remove_from_config,
    remove_from_configs_by_type,
    update_config,
)

__all__ = [
    "build_dataset",
    "build_dataloader",
    "build_data_parallel",
    "remove_from_config",
    "remove_from_configs_by_type",
    "get_configs_by_pairs",
    "get_configs_by_keys",
    "update_config",
    "get_dataset_configs",
    "prepare_for_testing",
    "is_epoch_based_runner",
    "config_from_string",
    "patch_default_config",
    "patch_data_pipeline",
    "patch_color_conversion",
    "patch_runner",
    "align_data_config_with_recipe",
    "get_meta_keys",
    "prepare_work_dir",
    "get_data_cfg",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmseg.models.losses.utils import weight_reduce_loss

from .base_pixel_loss import BasePixelLoss


class MPABasePixelLoss(BasePixelLoss):
    def __init__(self, **kwargs):
        super(MPABasePixelLoss, self).__init__(**kwargs)

    def _forward(
        self,
        output,
        labels,
        valid_label_mask,
        avg_factor=None,
        pixel_weights=None,
        reduction_override=None,
    ):
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = reduction_override if reduction_override else self.reduction

        self._last_scale = self._scale_scheduler(self.iter, self.epoch_size)

        if self.with_pr_product:
            output = self._pr_product(output)

        import numpy as np

        _labels = labels.cpu().detach().numpy()
        _labels[np.where((_labels == self.ignore_index))] = 0
        num_classes = output.size(1)
        valid_labels = torch.clamp(labels, 0, num_classes - 1)
        valid_mask = labels != self.ignore_index

        losses, updated_output = self._calculate(output, _labels, valid_label_mask, self._last_scale)

        if self.with_regularization:
            self._last_reg_weight = self._reg_weight_scheduler(self.iter, self.epoch_size)
            regularization = self._regularization(updated_output, self._last_scale, self._last_reg_weight)
            losses = torch.clamp_min(losses + regularization, 0.0)

        if self.with_border_reweighting:
            assert pixel_weights is not None
            losses = pixel_weights.squeeze(1) * losses

        losses = torch.where(valid_mask, losses, torch.zeros_like(losses))
        raw_sparsity = self._sparsity(losses, valid_mask)
        invalid_ratio = self._pred_stat(output, labels, valid_mask)

        weight, weight_sparsity = None, 0.0
        if self.sampler is not None:
            weight = self.sampler.sample(output, valid_labels, losses, valid_mask)
            weight_sparsity = self._sparsity(weight, valid_mask)

        loss = weight_reduce_loss(losses, weight=weight, reduction=reduction, avg_factor=avg_factor)

        meta = dict(
            weight=self.last_loss_weight,
            reg_weight=self.last_reg_weight,
            scale=self.last_scale,
            raw_sparsity=raw_sparsity,
            weight_sparsity=weight_sparsity,
            invalid_ratio=invalid_ratio,
        )

        return loss, meta



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

"""
The module with common logger for all OTX training tests.
"""

import logging


def get_logger():
    """
    The function returns the common logger for all OTX
    training tests.
    """
    logger_name = ".".join(__name__.split(".")[:-1])
    return logging.getLogger(logger_name)


def set_log_level(level, recursive=False):
    """
    The function sets log level for the common logger for all
    OTX training tests.
    The parameter `level` may be either int code or string
    (e.g. 'DEBUG')
    """

    logger = get_logger()
    logger.setLevel(level)
    if not recursive:
        return
    while logger:
        logger.setLevel(level)
        logger = logger.parent



"""Collection of video loading data pipelines for OTX Action Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict

import numpy as np
from mmaction.datasets.builder import PIPELINES

from otx.api.entities.datasets import DatasetEntity


@PIPELINES.register_module(force=True)
class RawFrameDecode:
    """Load and decode frames with given indices."""

    otx_dataset: DatasetEntity

    def __call__(self, results: Dict[str, Any]) -> Dict[str, Any]:
        """Call function of RawFrameDecode."""
        results = self._decode_from_list(results)
        return results

    def _decode_from_list(self, results: Dict[str, Any]):
        """Generate numpy array list from list of DatasetItemEntity."""
        imgs = []
        for index in results["frame_inds"]:
            imgs.append(self.otx_dataset[int(index)].media.numpy)
        results["imgs"] = imgs
        results["original_shape"] = imgs[0].shape[:2]
        results["img_shape"] = imgs[0].shape[:2]

        # we resize the gt_bboxes and proposals to their real scale
        if "gt_bboxes" in results:
            height, width = results["img_shape"]
            scale_factor = np.array([width, height, width, height])
            gt_bboxes = results["gt_bboxes"]
            gt_bboxes = (gt_bboxes * scale_factor).astype(np.float32)
            results["gt_bboxes"] = gt_bboxes
            if "proposals" in results and results["proposals"] is not None:
                proposals = results["proposals"]
                proposals = (proposals * scale_factor).astype(np.float32)
                results["proposals"] = proposals

        return results



"""MMDployment config of Resnet model for Instance-Seg Task."""

_base_ = ["../../base/deployments/base_instance_segmentation_dynamic.py"]

ir_config = dict(
    output_names=["boxes", "labels", "masks"],
)

backend_config = dict(
    # dynamic batch causes forever running openvino process
    model_inputs=[dict(opt_shapes=dict(input=[1, 3, 800, 1344]))],
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# cython: language_level=3

import cython
import numpy as np

cimport numpy as np

np.import_array()
import cv2
from PIL import Image

ctypedef np.int32_t INT32_t
ctypedef np.uint8_t UINT8_t


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray[INT32_t, ndim=1] c_histogram(const UINT8_t[:, :, :] image):
    cdef np.ndarray[INT32_t, ndim=1] hist = np.zeros((768,), dtype=np.int32)
    cdef INT32_t [:] hist_view = hist
    cdef int height, width, y, x
    cdef UINT8_t r, g, b

    height = image.shape[0]
    width = image.shape[1]

    for y in range(height):
        for x in range(width):
            hist_view[image[y][x][0] + 000] += 1
            hist_view[image[y][x][1] + 256] += 1
            hist_view[image[y][x][2] + 512] += 1

    return hist


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _c_lut(np.ndarray[UINT8_t, ndim=3] image, const UINT8_t[:] r_lut):
    assert np.PyArray_ISCONTIGUOUS(image), "Image should be contiguous."
    assert image.ndim == 3, "Image should be 3D array."
    assert image.shape[2] == 3, "Image should have 3 channels."

    cdef int h = image.shape[0]
    cdef int w = image.shape[1]
    cdef int x, y, curr_idx
    cdef int r, g, b
    cdef UINT8_t* data_ptr = &image[0,0,0]

    for y in range(h):
        for x in range(w):
            r = image[y, x, 0] + 000
            g = image[y, x, 1] + 256
            b = image[y, x, 2] + 512

            curr_idx = 3 * w * y + 3 * x

            data_ptr[curr_idx + 0] = r_lut[r]
            data_ptr[curr_idx + 1] = r_lut[g]
            data_ptr[curr_idx + 2] = r_lut[b]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def autocontrast(np.ndarray[UINT8_t, ndim=3] image, cutoff=0, ignore=None) -> np.ndarray:
    cdef int* h
    cdef int i, lo, hi, ix, cut, layer
    cdef double scale, offset
    cdef int[:] histogram
    cdef np.ndarray[UINT8_t, ndim=1] lut = np.zeros((768,), dtype=np.uint8)
    cdef UINT8_t[:] lut_view = lut
    cdef np.ndarray[UINT8_t, ndim=3] result

    histogram = c_histogram(image)

    for layer in range(0, 768, 256):
        h = &histogram[layer]

        if ignore is not None:
            # get rid of outliers
            try:
                h[ignore] = 0
            except TypeError:
                # assume sequence
                for ix in ignore:
                    h[ix] = 0
        if cutoff:
            # cut off pixels from both ends of the histogram
            if not isinstance(cutoff, tuple):
                cutoff = (cutoff, cutoff)
            # get number of pixels
            n = 0
            for ix in range(256):
                n = n + h[ix]
            # remove cutoff% pixels from the low end
            cut = n * cutoff[0] // 100
            for lo in range(256):
                if cut > h[lo]:
                    cut = cut - h[lo]
                    h[lo] = 0
                else:
                    h[lo] -= cut
                    cut = 0
                if cut <= 0:
                    break
            # remove cutoff% samples from the high end
            cut = n * cutoff[1] // 100
            for hi in range(255, -1, -1):
                if cut > h[hi]:
                    cut = cut - h[hi]
                    h[hi] = 0
                else:
                    h[hi] -= cut
                    cut = 0
                if cut <= 0:
                    break
        # find lowest/highest samples after preprocessing
        for lo in range(256):
            if h[lo]:
                break
        for hi in range(255, -1, -1):
            if h[hi]:
                break
        if hi <= lo:
            # don't bother
            for i in range(256):
                lut_view[layer + i] = i
        else:
            scale = 255.0 / (hi - lo)
            offset = -lo * scale
            for ix in range(256):
                i = ix
                ix = (int)(ix * scale + offset)
                if ix < 0:
                    ix = 0
                elif ix > 255:
                    ix = 255
                lut_view[layer + i] = ix

        layer += 256

    _c_lut(image, lut_view)
    return image



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os
import tempfile

import pytest

from otx.algorithms.detection.utils import generate_label_schema
from otx.algorithms.detection.utils.data import (
    find_label_by_name,
    format_list_to_str,
    get_anchor_boxes,
    get_sizes_from_dataset_entity,
    load_dataset_items_coco_format,
)
from otx.api.entities.label import Domain
from otx.api.entities.model_template import TaskType, task_type_to_label_domain
from otx.api.entities.subset import Subset
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    create_dummy_coco_json,
    generate_det_dataset,
)


@e2e_pytest_unit
@pytest.mark.parametrize("name", ["rectangle", "something"])
def test_find_label_by_name(name):
    classes = ("rectangle", "ellipse", "triangle")
    label_schema = generate_label_schema(classes, task_type_to_label_domain(TaskType.DETECTION))
    out = find_label_by_name(label_schema.get_labels(include_empty=False), name, Domain.DETECTION)
    assert out.name == name


@e2e_pytest_unit
def test_find_label_by_name_error():
    classes = ("rectangle", "rectangle", "triangle")
    label_schema = generate_label_schema(classes, task_type_to_label_domain(TaskType.DETECTION))
    with pytest.raises(ValueError):
        find_label_by_name(label_schema.get_labels(include_empty=False), "rectangle", Domain.DETECTION)


@e2e_pytest_unit
@pytest.mark.parametrize(
    "task_type, domain",
    [(TaskType.DETECTION, Domain.DETECTION), (TaskType.INSTANCE_SEGMENTATION, Domain.INSTANCE_SEGMENTATION)],
)
def test_load_dataset_items_coco_format(task_type, domain):
    _, labels = generate_det_dataset(task_type=task_type)
    tmp_dir = tempfile.TemporaryDirectory()
    with tempfile.TemporaryDirectory() as tmp_dir:
        fake_json_file = os.path.join(tmp_dir, "fake_data.json")
        create_dummy_coco_json(fake_json_file)
        data_root_dir = "./some_data_root_dir"
        with_mask = True if domain == Domain.INSTANCE_SEGMENTATION else False
        out = load_dataset_items_coco_format(
            fake_json_file,
            data_root_dir,
            subset=Subset.TRAINING,
            domain=domain,
            with_mask=with_mask,
            labels_list=labels,
        )
    assert out is not None


@e2e_pytest_unit
def test_get_sizes_from_dataset_entity():
    dataset, _ = generate_det_dataset(task_type=TaskType.DETECTION)
    out = get_sizes_from_dataset_entity(dataset, [480, 640])
    assert out is not None


@e2e_pytest_unit
def test_get_anchor_boxes():
    out = get_anchor_boxes([(100, 120), (100, 120)], [1, 1])
    expected_out = ([[100.0], [100.0]], [[120.0], [120.0]])
    assert out == expected_out


@e2e_pytest_unit
def test_format_list_to_str():
    out = format_list_to_str([[0.1839128319, 0.47398123]])
    expected_out = "[[0.18, 0.47]]"
    assert out == expected_out



_base_ = "./schedule.py"

lr_config = dict(
    policy="ReduceLROnPlateau",
    metric="bbox_mAP",
    patience=30,
    iteration_patience=0,
    interval=1,
    min_lr=1e-06,
    warmup="linear",
    warmup_iters=200,
    warmup_ratio=0.3333333333333333,
)



"""Evaluation metrics."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .accuracy import Accuracy
from .averaging import MetricAverageMethod
from .basic_operations import (
    get_intersections_and_cardinalities,
    intersection_box,
    intersection_over_union,
    precision_per_class,
    recall_per_class,
)
from .dice import DiceAverage
from .f_measure import FMeasure
from .metrics_helper import MetricsHelper

__all__ = [
    "Accuracy",
    "MetricAverageMethod",
    "DiceAverage",
    "FMeasure",
    "intersection_box",
    "intersection_over_union",
    "MetricsHelper",
    "precision_per_class",
    "recall_per_class",
    "get_intersections_and_cardinalities",
]



from typing import List
from unittest.mock import MagicMock

import numpy as np
import pytest
from mmcv.runner import BaseRunner

from otx.algorithms.common.adapters.mmcv.hooks import (
    EMAMomentumUpdateHook,
    TwoCropTransformHook,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@pytest.fixture
def mock_iter_runner(mocker):
    _mock_iter_runner = mocker.patch("mmcv.runner.IterBasedRunner", autospec=True)
    _mock_iter_runner.data_loader = MagicMock()
    _mock_iter_runner.model = MagicMock()

    return _mock_iter_runner


@pytest.fixture
def mock_epoch_runner(mocker):
    _mock_epoch_runner = mocker.patch("mmcv.runner.EpochBasedRunner", autospec=True)
    _mock_epoch_runner.model = MagicMock()

    return _mock_epoch_runner


@pytest.mark.usefixtures("mock_epoch_runner")
@pytest.mark.usefixtures("mock_iter_runner")
class TestEMAMomentumUpdateHook:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.ema_momentum_update_hook = EMAMomentumUpdateHook()

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "by_epoch,cur_epoch,expected",
        [
            (True, 1, 0.9960002467350366),
            (True, 50, 0.9965857864376269),
            (True, 150, 0.9994142135623731),
            (False, 200, 0.996),
        ],
    )
    def test_before_train_epoch(
        self, mock_epoch_runner: BaseRunner, by_epoch: bool, cur_epoch: int, expected: float
    ) -> None:
        """Test before_train_epoch."""
        mock_epoch_runner.model.momentum = 0.996  # default value of BYOL
        mock_epoch_runner.model.base_momentum = 0.996
        mock_epoch_runner.epoch = cur_epoch
        mock_epoch_runner.max_epochs = 200
        self.ema_momentum_update_hook.by_epoch = by_epoch

        self.ema_momentum_update_hook.before_train_epoch(mock_epoch_runner)

        assert np.allclose(mock_epoch_runner.model.momentum, expected)

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "by_epoch,cur_iter,expected",
        [
            (False, 1, 0.9960002467350366),
            (False, 50, 0.9965857864376269),
            (False, 150, 0.9994142135623731),
            (True, 200, 0.996),
        ],
    )
    def test_before_train_iter(
        self, mock_iter_runner: BaseRunner, by_epoch: bool, cur_iter: int, expected: float
    ) -> None:
        """Test before_train_iter."""
        mock_iter_runner.model.momentum = 0.996  # default value of BYOL
        mock_iter_runner.model.base_momentum = 0.996
        mock_iter_runner.iter = cur_iter
        mock_iter_runner.max_iters = 200
        self.ema_momentum_update_hook.by_epoch = by_epoch

        self.ema_momentum_update_hook.before_train_iter(mock_iter_runner)

        assert np.allclose(mock_iter_runner.model.momentum, expected)

    @e2e_pytest_unit
    def test_after_train_iter(self, mock_iter_runner: BaseRunner) -> None:
        """Test after_train_iter."""
        mock_iter_runner.iter = 1
        mock_iter_runner.model.momentum_update = MagicMock()

        self.ema_momentum_update_hook.after_train_iter(mock_iter_runner)

        mock_iter_runner.model.momentum_update.assert_called_once()


@pytest.mark.usefixtures("mock_iter_runner")
class TestTwoCropTransformHook:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.two_crop_transform_hook = TwoCropTransformHook(interval=1)

    def set_mock_name(self, name: str, is_both: bool = True) -> MagicMock:
        mock_class = MagicMock()
        mock_class.__class__.__name__ = name
        if name == "TwoCropTransform":
            mock_class.is_both = is_both
        return mock_class

    @e2e_pytest_unit
    def test_use_not_implemented_by_epoch(self) -> None:
        with pytest.raises(NotImplementedError):
            TwoCropTransformHook(interval=1, by_epoch=True)

    @e2e_pytest_unit
    def test_get_dataset(self, mock_iter_runner: BaseRunner) -> None:
        """Test _get_dataset."""
        mock_iter_runner.data_loader.dataset = True
        assert not hasattr(mock_iter_runner.data_loader.dataset, "dataset")

        results = self.two_crop_transform_hook._get_dataset(mock_iter_runner)
        assert results

    @e2e_pytest_unit
    def test_get_dataset_repeat_dataset(self, mock_iter_runner: BaseRunner) -> None:
        """Test _get_dataset when dataset includes child dataset (e.g. RepeatDataset)."""
        mock_iter_runner.data_loader.dataset.dataset = True
        assert hasattr(mock_iter_runner.data_loader.dataset, "dataset")

        results = self.two_crop_transform_hook._get_dataset(mock_iter_runner)
        assert results

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "transforms_order",
        [["TwoCropTransform", "A", "B"], ["A", "TwoCropTransform", "B"], ["A", "B", "TwoCropTransform"]],
    )
    def test_find_two_crop_transform(self, transforms_order: List[object]) -> None:
        """Test _find_two_crop_transform."""
        transforms = [self.set_mock_name(name) for name in transforms_order]

        results = self.two_crop_transform_hook._find_two_crop_transform(transforms)
        assert results.__class__.__name__ == "TwoCropTransform"

    @e2e_pytest_unit
    @pytest.mark.parametrize("interval,cnt,expected", [(1, 0, True), (2, 0, False), (2, 1, True)])
    def test_before_train_epoch(self, mock_iter_runner: BaseRunner, interval: int, cnt: int, expected: bool) -> None:
        """Test before_train_epoch."""
        if hasattr(mock_iter_runner.data_loader.dataset, "dataset"):
            del mock_iter_runner.data_loader.dataset.dataset

        setattr(self.two_crop_transform_hook, "interval", interval)
        setattr(self.two_crop_transform_hook, "cnt", cnt)

        transforms_order = ["TwoCropTransform", "A", "B"]
        transforms = [self.set_mock_name(name=name) for name in transforms_order]
        mock_iter_runner.data_loader.dataset.pipeline.transforms = transforms

        self.two_crop_transform_hook.before_train_epoch(mock_iter_runner)

        assert transforms[0].is_both == expected

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "is_both,interval,cnt,expected",
        [
            (False, 1, 0, False),
            (True, 1, 0, True),
            (False, 3, 0, False),
            (True, 3, 0, True),
            (False, 3, 1, True),
            (True, 3, 1, False),
            (True, 2, 0, False),
        ],
    )
    def test_after_train_iter(
        self, mock_iter_runner: BaseRunner, is_both: bool, interval: int, cnt: int, expected: bool
    ) -> None:
        """Test after_train_iter."""
        if hasattr(mock_iter_runner.data_loader.dataset, "dataset"):
            del mock_iter_runner.data_loader.dataset.dataset

        setattr(self.two_crop_transform_hook, "interval", interval)
        setattr(self.two_crop_transform_hook, "cnt", cnt)

        transforms_order = ["TwoCropTransform", "A", "B"]
        transforms = [self.set_mock_name(name=name, is_both=is_both) for name in transforms_order]
        mock_iter_runner.data_loader.dataset.pipeline.transforms = transforms

        self.two_crop_transform_hook.after_train_iter(mock_iter_runner)

        assert transforms[0].is_both == expected
        if is_both and interval - cnt == 2:
            # test cnt initialization
            assert self.two_crop_transform_hook.cnt == 0



"""Train Task of OTX Action Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import os
from glob import glob
from typing import Any, Iterable, List, Optional

import numpy as np
import torch
from mmaction.apis import train_model
from mmaction.datasets import build_dataset
from mmaction.utils import get_root_logger

from otx.algorithms.action.adapters.mmaction.utils import prepare_for_training
from otx.algorithms.common.utils import TrainingProgressCallback
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    ScoreMetric,
    VisualizationType,
)
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.train_parameters import TrainParameters, default_progress_callback
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.tasks.interfaces.training_interface import ITrainingTask

from .inference import ActionInferenceTask

logger = get_root_logger()


# pylint: disable=too-many-locals, too-many-instance-attributes
class ActionTrainTask(ActionInferenceTask, ITrainingTask):
    """Train Task Implementation of OTX Action Task."""

    def save_model(self, output_model: ModelEntity):
        """Save best model weights in ActionTrainTask."""
        logger.info("called save_model")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt["state_dict"],
            "config": hyperparams_str,
            "labels": labels,
            "confidence_threshold": self.confidence_threshold,
            "VERSION": 1,
        }

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision

    def cancel_training(self):
        """Cancel training function in ActionTrainTask.

        Sends a cancel training signal to gracefully stop the optimizer. The signal consists of creating a
        '.stop_training' file in the current work_dir. The runner checks for this file periodically.
        The stopping mechanism allows stopping after each iteration, but validation will still be carried out. Stopping
        will therefore take some time.
        """
        logger.info("Cancel training requested.")
        self._should_stop = True
        if self.cancel_interface is not None:
            self.cancel_interface.cancel()
        else:
            logger.info("but training was not started yet. reserved it to cancel")
            self.reserved_cancel = True

    def train(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        train_parameters: Optional[TrainParameters] = None,
    ):
        """Train function in ActionTrainTask."""
        logger.info("train()")
        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # Set OTE LoggerHook & Time Monitor
        if train_parameters:
            update_progress_callback = train_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback
        self._time_monitor = TrainingProgressCallback(update_progress_callback)

        self._is_training = True
        self._init_task()

        if self._recipe_cfg is None:
            raise Exception("Recipe config is not initialized properly")

        results = self._train_model(dataset)

        # Check for stop signal when training has stopped. If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        self._get_output_model(results)
        performance = self._get_final_eval_results(dataset, output_model)

        # save resulting model
        self.save_model(output_model)
        output_model.performance = performance
        self._is_training = False
        logger.info("train done.")

    def _train_model(self, dataset: DatasetEntity):
        if self._recipe_cfg is None:
            raise Exception("Recipe config does not initialize properly!")
        train_dataset = dataset.get_subset(Subset.TRAINING)
        val_dataset = dataset.get_subset(Subset.VALIDATION)
        training_config = prepare_for_training(
            self._recipe_cfg, train_dataset, val_dataset, self._time_monitor, self._learning_curves
        )
        mm_train_dataset = build_dataset(training_config.data.train)
        logger.info("Start training")
        self._model.train()
        # FIXME runner is built inside of train_model funciton, it is hard to change runner's type
        train_model(model=self._model, dataset=mm_train_dataset, cfg=training_config, validate=True)
        checkpoint_file_path = glob(os.path.join(self._recipe_cfg.work_dir, "best*pth"))
        if len(checkpoint_file_path) == 0:
            checkpoint_file_path = os.path.join(self._recipe_cfg.work_dir, "latest.pth")
        elif len(checkpoint_file_path) > 1:
            logger.warning(f"Multiple candidates for the best checkpoint found: {checkpoint_file_path}")
            checkpoint_file_path = checkpoint_file_path[0]
        else:
            checkpoint_file_path = checkpoint_file_path[0]
        logger.info(f"Use {checkpoint_file_path} for final model weights")

        return {"final_ckpt": checkpoint_file_path}

    def _get_output_model(self, results):
        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt
        self._model.load_state_dict(torch.load(self._model_ckpt)["state_dict"])

    def _get_final_eval_results(self, dataset, output_model):
        logger.info("Final Evaluation")
        val_dataset = dataset.get_subset(Subset.VALIDATION)
        val_preds, val_map = self._infer_model(val_dataset, InferenceParameters(is_evaluation=True))

        preds_val_dataset = val_dataset.with_empty_annotations()
        # TODO Load _add_predictions_to_dataset function from self._task_type
        if self._task_type == TaskType.ACTION_CLASSIFICATION:
            self._add_predictions_to_dataset(val_preds, preds_val_dataset)
        elif self._task_type == TaskType.ACTION_DETECTION:
            self._add_det_predictions_to_dataset(val_preds, preds_val_dataset)

        result_set = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=val_dataset,
            prediction_dataset=preds_val_dataset,
        )

        metric = self._get_metric(result_set)

        # compose performance statistics
        performance = metric.get_performance()
        metric_name = self._recipe_cfg.evaluation.final_metric
        performance.dashboard_metrics.extend(
            ActionTrainTask._generate_training_metrics(self._learning_curves, val_map, metric_name)
        )
        logger.info(f"Final model performance: {str(performance)}")
        return performance

    @staticmethod
    # TODO Implement proper function for action classification
    def _generate_training_metrics(learning_curves, scores, metric_name) -> Iterable[MetricsGroup[Any, Any]]:
        """Get Training metrics (epochs & scores).

        Parses the mmaction logs to get metrics from the latest training run
        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []

        # Learning curves.
        for key, curve in learning_curves.items():
            len_x, len_y = len(curve.x), len(curve.y)
            if len_x != len_y:
                logger.warning(f"Learning curve {key} has inconsistent number of coordinates ({len_x} vs {len_y}.")
                len_x = min(len_x, len_y)
                curve.x = curve.x[:len_x]
                curve.y = curve.y[:len_x]
            metric_curve = CurveMetric(
                xs=np.nan_to_num(curve.x).tolist(),
                ys=np.nan_to_num(curve.y).tolist(),
                name=key,
            )
            visualization_info = LineChartInfo(name=key, x_axis_label="Epoch", y_axis_label=key)
            output.append(LineMetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))

        # Final mAP value on the validation set.
        output.append(
            BarMetricsGroup(
                metrics=[ScoreMetric(value=scores, name=f"{metric_name}")],
                visualization_info=BarChartInfo("Validation score", visualization_type=VisualizationType.RADIAL_BAR),
            )
        )

        return output



"""Tests for OTX CLI commands"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import pytest

from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    otx_build_auto_config,
    otx_build_backbone_testing,
    otx_find_testing,
    otx_train_auto_config,
)

otx_dir = os.getcwd()


build_backbone_args = [
    ("CLASSIFICATION", "torchvision.mobilenet_v3_large"),
    ("CLASSIFICATION", "mmcls.MMOVBackbone"),
    ("DETECTION", "torchvision.mobilenet_v3_large"),
    ("INSTANCE_SEGMENTATION", "torchvision.mobilenet_v3_large"),
    ("SEGMENTATION", "torchvision.mobilenet_v3_large"),
]
build_backbone_args_ids = [f"{task}_{backbone}" for task, backbone in build_backbone_args]


class TestToolsOTXCLI:
    @e2e_pytest_component
    def test_otx_find(self):
        otx_find_testing()

    @e2e_pytest_component
    @pytest.mark.parametrize("build_backbone_args", build_backbone_args, ids=build_backbone_args_ids)
    def test_otx_backbone_build(self, tmp_dir_path, build_backbone_args):
        tmp_dir_path = tmp_dir_path / build_backbone_args[0] / build_backbone_args[1]
        otx_build_backbone_testing(tmp_dir_path, build_backbone_args)


auto_config_args_with_autosplit = {"--train-data-roots": "tests/assets/imagenet_dataset"}

auto_config_args_with_autosplit_task = {
    "--task": "classification",
    "--train-data-roots": "tests/assets/imagenet_dataset",
}

auto_config_args_without_autosplit = {
    "--train-data-roots": "tests/assets/imagenet_dataset",
    "--val-data-roots": "tests/assets/imagenet_dataset_class_incremental",
}

build_auto_config_args = {
    "classification": {"--train-data-roots": "tests/assets/imagenet_dataset"},
    "classification_with_task": {"--task": "classification", "--train-data-roots": "tests/assets/imagenet_dataset"},
    "detection": {"--train-data-roots": "tests/assets/car_tree_bug"},
    "detection_with_task": {"--task": "detection", "--train-data-roots": "tests/assets/car_tree_bug"},
}


class TestToolsOTXBuildAutoConfig:
    @e2e_pytest_component
    @pytest.mark.parametrize("case", build_auto_config_args.keys())
    def test_otx_build_with_autosplit(self, case, tmp_dir_path):
        otx_dir = os.getcwd()
        tmp_dir_path = tmp_dir_path / "test_build_auto_config" / case
        otx_build_auto_config(root=tmp_dir_path, otx_dir=otx_dir, args=build_auto_config_args[case])


train_auto_config_args = {
    "classification": {"--train-data-roots": "tests/assets/imagenet_dataset"},
    "classification_with_template": {
        "template": "otx/algorithms/classification/configs/mobilenet_v3_large_1_cls_incr/template.yaml",
        "--train-data-roots": "tests/assets/imagenet_dataset",
    },
    "detection": {"--train-data-roots": "tests/assets/car_tree_bug"},
    "detection_with_template": {
        "template": "otx/algorithms/detection/configs/detection/mobilenetv2_ssd/template.yaml",
        "--train-data-roots": "tests/assets/car_tree_bug",
    },
}

train_params = [
    "params",
    "--learning_parameters.num_iters",
    "4",
]


class TestToolsOTXTrainAutoConfig:
    @e2e_pytest_component
    @pytest.mark.parametrize("case", train_auto_config_args.keys())
    def test_otx_train(self, case, tmp_dir_path):
        otx_dir = os.getcwd()
        tmp_dir_path = tmp_dir_path / case
        train_auto_config_args[case]["train_params"] = train_params
        otx_train_auto_config(root=tmp_dir_path, otx_dir=otx_dir, args=train_auto_config_args[case])
